from __future__ import annotations

import itertools
from collections.abc import (
    Iterable,
)
from heapq import heapify, heappop, heappush
from typing import Optional, Any, Generic, Union

from ._gear_collections import (
    access_to_vertex_set,
    access_to_vertex_mapping,
    access_to_vertex_mapping_expect_none,
)
from ._gears import (
    GearWithoutDistances,
    Gear,
    GearDefault,
)
from ._path import (
    Path,
    PathOfUnlabeledEdges,
    PathOfLabeledEdges,
)
from ._strategies import (
    Strategy,
    T_strategy,
    BNextVertices,
    BNextEdges,
    BNextLabeledEdges,
    BNextWeightedEdges,
    BNextWeightedLabeledEdges,
    BNextEdgesOrVertices,
    BNextWeightedMaybeLabeledEdges,
    iter_start_vertices_and_ids,
    define_distances,
    create_paths,
)
from ._traversals import TraversalBreadthFirstFlex
from ._types import (
    T_vertex,
    T_vertex_id,
    T_weight,
    T_labels,
    VertexToID,
    vertex_as_id,
)


# --------------- internal support functions -------------


def _search_needs_search_object(obj: Any, needed_class: type) -> None:
    if not isinstance(obj, needed_class):
        raise RuntimeError(
            "Method start_from can only be called on a search strategy object."
        )


def _create_unified_next_bidirectional(
    next_vertices: Optional[BNextVertices[T_vertex, T_strategy]],
    next_edges: Optional[BNextEdges[T_vertex, T_strategy]],
    next_labeled_edges: Optional[BNextLabeledEdges[T_vertex, T_strategy, T_labels]],
) -> tuple[BNextEdgesOrVertices[T_vertex, T_strategy, T_labels], bool, bool]:
    """Check configuration of given next_vertices, next_edges, and next_labeled_edges
    function pairs
    and calculate a pair of unified NextEdgesOrVertices
    and whether we have edges with data (weights and/or labels) and/or labeled_edges.
    """
    next_edges_or_vertices: BNextEdgesOrVertices[T_vertex, T_strategy, T_labels]
    if next_vertices is not None:
        if next_edges is not None:
            raise RuntimeError("Both next_vertices and next_edges provided.")
        if next_labeled_edges is not None:
            raise RuntimeError("Both next_vertices and next_labeled_edges provided.")
        next_edges_or_vertices = next_vertices
        edges_with_data = False
        labeled_edges = False
    elif next_edges is not None:
        if next_labeled_edges is not None:
            raise RuntimeError("Both next_edges and next_labeled_edges provided.")
        next_edges_or_vertices = next_edges
        edges_with_data = True
        labeled_edges = False
    else:
        if next_labeled_edges is None:
            raise RuntimeError(
                "Neither next_vertices nor next_edges "
                + "nor next_labeled_edges provided."
            )
        next_edges_or_vertices = next_labeled_edges
        edges_with_data = True
        labeled_edges = True
    return next_edges_or_vertices, edges_with_data, labeled_edges


def _create_unified_next_weighted_bidirectional(
    next_edges: Optional[BNextWeightedEdges[T_vertex, T_strategy, T_weight]],
    next_labeled_edges: Optional[
        BNextWeightedLabeledEdges[T_vertex, T_strategy, T_weight, T_labels]
    ],
) -> tuple[
    BNextWeightedMaybeLabeledEdges[T_vertex, T_strategy, T_weight, T_labels],
    bool,
]:
    """Check configuration of given next_edges and next_labeled_edges function pairs
    and calculate a pair of unified _NextWeightedMaybeLabeledEdges and whether we have
    labeled_edges.
    """
    next_maybe_labeled_edges: BNextWeightedMaybeLabeledEdges[
        T_vertex, T_strategy, T_weight, T_labels
    ]
    if next_edges is not None:
        if next_labeled_edges is not None:
            raise RuntimeError("Both next_edges and next_labeled_edges provided.")
        next_maybe_labeled_edges = next_edges
        labeled_edges = False
    else:
        if next_labeled_edges is None:
            raise RuntimeError("Neither next_edges and next_labeled_edges provided.")
        next_maybe_labeled_edges = next_labeled_edges
        labeled_edges = True
    return next_maybe_labeled_edges, labeled_edges


# ------------------------- Bidirectional search -----------------------


class BSearchBreadthFirstFlex(Strategy[T_vertex, T_vertex_id, T_labels]):
    """
    Bases: `Strategy` [`T_vertex`, `T_vertex_id`, `T_labels`]

    :param vertex_to_id: See `VertexToID` function.

    :param gear: See `gears API <gears_api>` and class `GearWithoutDistances`.

    :param next_vertices: Tuple `BNextVertices` of two NextVertices function.
      If None, provide next_edges or next_labeled_edges.

    :param next_edges: Tuple `BNextEdges` of two NextEdges functions.
      See paragraph *input* below for details. Only allowed if next_vertices equals
      None. If both are None, provide next_labeled_edges.

    :param next_labeled_edges: Tuple `BNextLabeledEdges` of two NextEdges
      functions. See paragraph *input* below for details. The parameter is only
      allowed if next_vertices and next_edges equal None. If given, paths will record
      the given labels.

    **Algorithm:** Bidirectional version of the Breadth First Search algorithm,
    non-recursive, based on FIFO queues.

    **Properties:** In both directions, vertices are visited by increasing depth
    from a start (resp. a goal) vertex (minimal number of edges), till a shortest
    path (minimal number of edges) from a start to a goal vertex is found. Each
    vertex is visited only once.

    **Input:** Directed graph. Unlabeled or labeled edges. One or more start vertices,
    and one or more goal vertices. NextVertices (resp. NextEdges) functions both for
    the outgoing edges from a vertex and the incoming edges to a vertex have to be
    provided, and they need to describe the same graph. Optional calculation limit.

    Note: A shortest path from a vertex *v* to itself always exists, has an edge count
    of 0, and will be found by the class, whilst TraversalBreadthFirst does not report
    start vertices and thus,
    TraversalBreadthFirst(<something>).start_at(v).go_to(v) fails.
    """

    def __init__(
        self,
        vertex_to_id: VertexToID[T_vertex, T_vertex_id],
        gear: GearWithoutDistances[T_vertex, T_vertex_id, T_labels],
        next_vertices: Optional[
            BNextVertices[
                T_vertex, TraversalBreadthFirstFlex[T_vertex, T_vertex_id, T_labels]
            ]
        ] = None,
        *,
        next_edges: Optional[
            BNextEdges[
                T_vertex,
                TraversalBreadthFirstFlex[T_vertex, T_vertex_id, T_labels],
            ],
        ] = None,
        next_labeled_edges: Optional[
            BNextLabeledEdges[
                T_vertex,
                TraversalBreadthFirstFlex[T_vertex, T_vertex_id, T_labels],
                T_labels,
            ],
        ] = None,
    ) -> None:
        self._vertex_to_id = vertex_to_id
        self._gear = gear

        (
            _,
            self._edges_with_data,
            self._labeled_edges,
        ) = _create_unified_next_bidirectional(
            next_vertices, next_edges, next_labeled_edges
        )

        self._traversal_bi = tuple(
            TraversalBreadthFirstFlex[T_vertex, T_vertex_id, T_labels](
                vertex_to_id,
                gear,
                next_vertices=None if next_vertices is None else next_vertices[i],
                next_edges=None if next_edges is None else next_edges[i],
                next_labeled_edges=(
                    None if next_labeled_edges is None else next_labeled_edges[i]
                ),
            )
            for i in range(2)
        )

    def start_from(
        self,
        start_and_goal_vertex: Optional[tuple[T_vertex, T_vertex]] = None,
        *,
        start_and_goal_vertices: Optional[
            tuple[Iterable[T_vertex], Iterable[T_vertex]]
        ] = None,
        build_path: bool = False,
        calculation_limit: Optional[int] = None,
        fail_silently: bool = False,
    ) -> tuple[int, Path[T_vertex, T_vertex_id, T_labels]]:
        """
        Start the search both from a start vertex and a goal vertex, resp. both
        from a set of start vertices and a set of goal vertices. Return the
        length of a shortest (sum of edge weights) path between the/a start vertex
        and the/a goal vertex. If building a path was requested, also return the path,
        and otherwise, return a dummy path object.

        If the search ends without having found a path, raise KeyError,
        or, if a silent fail is demanded, return -1 and a dummy path object.
        Here, infinity means the value for infinite distance that is defined by the
        used `Gear` (as provided by a call of gear.infinity()).

        NoGraphs gives no guarantees about a dummy path. Do not use it for anything.

        :param start_and_goal_vertex: The start vertex and the goal vertex of the
            search. If None, provide start_and_goal_vertices.

        :param start_and_goal_vertices: The start vertices and the
            goal vertices of the search. Only allowed if start_and_goal_vertex
            equals None.

        :param build_path: If true, build and return a path of the minimum possible
            length.

        :param calculation_limit: If provided, maximal number of vertices to process
            (read in) from your graph in each of the searches in one of the two
            directions. If it is exceeded, a RuntimeError will be raised.

        :param fail_silently: If no path can be found, fail silently (see above)
            instead of raising an exception.
        """
        # steps that other strategies do in start_from

        _search_needs_search_object(self, BSearchBreadthFirstFlex)

        start_vertices: Iterable[T_vertex]
        goal_vertices: Iterable[T_vertex]
        if start_and_goal_vertex is not None:
            if start_and_goal_vertices is not None:
                raise RuntimeError(
                    "Both start_and_goal_vertex and "
                    + "start_and_goal_vertices provided."
                )
            start_vertex, goal_vertex = start_and_goal_vertex
            start_vertices, goal_vertices = (start_vertex,), (goal_vertex,)
        else:
            if start_and_goal_vertices is None:
                raise RuntimeError(
                    "Neither start_and_goal_vertex nor "
                    + "start_and_goal_vertices provided."
                )
            start_vertices, goal_vertices = start_and_goal_vertices

        for t, vertices in zip(self._traversal_bi, (start_vertices, goal_vertices)):
            t.start_from(
                start_vertices=vertices,
                build_paths=build_path,
                calculation_limit=calculation_limit,
                _report_depth_increase=True,
            )

        visited_bi = tuple(t.visited for t in self._traversal_bi)

        # Copy Traversal attributes into method scope (faster access)
        # labeled_edges = self._labeled_edges
        maybe_vertex_to_id = (
            None if self._vertex_to_id == vertex_as_id else self._vertex_to_id
        )  # Case vertex_as_id: not apply; T_vertex_id > T_vertex

        # Get references of used gear objects and methods (avoid attribute resolution)
        (
            visited_forwards_uses_sequence,
            visited_forwards_sequence,
            visited_forwards_wrapper,
            visited_forwards_uses_bits,
            visited_forwards_index_and_bit_method,
        ) = access_to_vertex_set(visited_bi[0])
        (
            visited_backwards_uses_sequence,
            visited_backwards_sequence,
            visited_backwards_wrapper,
            visited_backwards_uses_bits,
            visited_backwards_index_and_bit_method,
        ) = access_to_vertex_set(visited_bi[1])

        # ----- Initialize method specific bookkeeping -----

        # Get the right class for storing a path (labeled or not)
        path_cls: type[Path]
        if self._labeled_edges:
            path_cls = PathOfLabeledEdges[T_vertex, T_vertex_id, T_labels]
        else:
            path_cls = PathOfUnlabeledEdges[T_vertex, T_vertex_id, T_labels]

        # Detect if a start vertex is also goal vertex, and report result manually.
        # (Without this manual handling, a non-self loop from such a start vertex back
        #  to itself with a length > 0 would be reported as smallest distance. This
        #  would be unexpected for users, since they expect that the distance from a
        #  vertex to itself is always 0.)
        common_vertex_ids = set(
            v_id
            for v, v_id in iter_start_vertices_and_ids(
                start_vertices, self._vertex_to_id
            )
        ).intersection(
            v_id
            for v, v_id in iter_start_vertices_and_ids(
                goal_vertices, self._vertex_to_id
            )
        )
        if common_vertex_ids:
            for c_vertex, cv_id in iter_start_vertices_and_ids(
                start_vertices, self._vertex_to_id
            ):
                if cv_id in common_vertex_ids:
                    p = path_cls.from_vertex(c_vertex)
                    return 0, p

        # ----- Inner loop -----

        for (
            traversal_iter,
            visited_other,
            visited_other_uses_sequence,
            visited_other_sequence,
            visited_other_wrapper,
            visited_other_uses_bits,
            visited_other_index_and_bit_method,
            paths,
        ) in itertools.cycle(
            (
                (
                    iter(self._traversal_bi[0]),
                    visited_bi[1],
                    visited_backwards_uses_sequence,
                    visited_backwards_sequence,
                    visited_backwards_wrapper,
                    visited_backwards_uses_bits,
                    visited_backwards_index_and_bit_method,
                    self._traversal_bi[0].paths,
                ),
                (
                    iter(self._traversal_bi[1]),
                    visited_bi[0],
                    visited_forwards_uses_sequence,
                    visited_forwards_sequence,
                    visited_forwards_wrapper,
                    visited_forwards_uses_bits,
                    visited_forwards_index_and_bit_method,
                    self._traversal_bi[1].paths,
                ),
            )
        ):
            prev_vertex: Optional[T_vertex] = None
            for vertex in traversal_iter:
                # If we get the same vertex twice, directly after each other,
                # this signals, that the depth will increase with the next reported
                # vertex. So we leave the loop here.
                if prev_vertex == vertex:
                    break
                prev_vertex = vertex

                # If vertex is not in visited vertices of other traversal: continue
                v_id: T_vertex_id = (
                    maybe_vertex_to_id(vertex)  # type: ignore[assignment]
                    if maybe_vertex_to_id
                    else vertex
                )

                if not visited_other_uses_sequence:
                    # Standard implementation for "normal" MutableMapping
                    if v_id not in visited_other:
                        continue
                elif visited_other_uses_bits:
                    # Same as above, but with bits in byte sequence
                    sequence_key, bit_number = visited_other_index_and_bit_method(
                        v_id, 8
                    )
                    bit_mask = 1 << bit_number
                    try:
                        value = visited_other_sequence[sequence_key]
                        if not (value & bit_mask):
                            continue
                    except IndexError:
                        continue
                else:
                    # Same as above, but with booleans in byte sequence
                    try:
                        if not visited_other_sequence[v_id]:
                            continue
                    except IndexError:
                        continue

                # We found a vertex from both directions
                path = (
                    path_cls.from_bidirectional_search(
                        self._traversal_bi[0].paths, self._traversal_bi[1].paths, vertex
                    )
                    if build_path
                    else path_cls.of_nothing()
                )
                return sum(t.depth for t in self._traversal_bi), path
            else:
                # No new vertices reported by traversal in this direction and depth:
                # Whole search is over.
                break

        if fail_silently:
            return -1, path_cls.of_nothing()
        else:
            raise KeyError("No path to (a) goal vertex found")


class BSearchBreadthFirst(
    Generic[T_vertex, T_labels], BSearchBreadthFirstFlex[T_vertex, T_vertex, T_labels]
):
    """
    Eases the use of `BSearchBreadthFirstFlex` for typical cases.
    For documentation of functionality and parameters, see there.

    .. code-block:: python

       BSearchBreadthFirst[T_vertex, T_labels](*args, **keywords)

    is a short form for

    .. code-block:: python

       BSearchBreadthFirstFlex[
           T_vertex, T_vertex, T_labels],
      ](nog.vertex_as_id, nog.GearDefault(), *args, **keywords)

    Implication:

    - `GearDefault` is used, see there how it and its superclass work
    - T_vertex is bound to Hashable (T_vertex is used as `T_vertex_id`, see there)
    """

    def __init__(
        self,
        next_vertices: Optional[
            BNextVertices[
                T_vertex, TraversalBreadthFirstFlex[T_vertex, T_vertex, T_labels]
            ]
        ] = None,
        *,
        next_edges: Optional[
            BNextEdges[
                T_vertex,
                TraversalBreadthFirstFlex[T_vertex, T_vertex, T_labels],
            ],
        ] = None,
        next_labeled_edges: Optional[
            BNextLabeledEdges[
                T_vertex,
                TraversalBreadthFirstFlex[T_vertex, T_vertex, T_labels],
                T_labels,
            ],
        ] = None,
    ) -> None:
        super().__init__(
            vertex_as_id,
            GearDefault(),
            next_vertices,
            next_edges=next_edges,
            next_labeled_edges=next_labeled_edges,
        )


class BSearchShortestPathFlex(
    Strategy[T_vertex, T_vertex_id, T_labels],
    Generic[T_vertex, T_vertex_id, T_weight, T_labels],
):
    """
    Bases: `Strategy` [`T_vertex`, `T_vertex_id`, `T_labels`],
    Generic[`T_vertex`, `T_vertex_id`, `T_weight`, `T_labels`]

    :param vertex_to_id: See `VertexToID` function.

    :param gear: See `gears API <gears_api>` and class `Gear`.

    :param next_edges: Tuple `BNextWeightedEdges` of two NextEdges functions.
      See paragraph *input* below for details. If None, provide next_labeled_edges.

    :param next_labeled_edges: Tuple `BNextWeightedLabeledEdges` of two NextEdges
      functions. See paragraph *input* below for details. The parameter is only
      allowed if next_edges equals None. If given, paths will record the given labels.

    **Algorithm:** Bidirectional version of the shortest paths algorithm of Dijkstra,
    non-recursive, based on heap.

    **Properties:** In both directions, vertices are visited by increasing distance
    (minimally necessary sum of edge weights) from a start (resp. a goal) vertex,
    till a shortest path (minimal sum of edge weights) from a start to a goal vertex
    is found. Each vertex is visited only once.

    **Input:** Weighted directed graph. One or more start vertices, and one or more
    goal vertices. Weights need to be non-negative. NextWeightedEdges (resp.
    NextWeightedLabeledEdges) functions both for the outgoing edges from a vertex
    and the incoming edges to a vertex have to be provided, and they need to describe
    the same graph. Optional calculation limit.

    Note: A shortest path from a vertex *v* to itself always exists, has length 0,
    and will be found by the class, whilst TraversalShortestPaths does not report
    start vertices and thus,
    TraversalShortestPaths(<something>).start_at(v).go_to(v) fails.
    """

    def __init__(
        self,
        vertex_to_id: VertexToID[T_vertex, T_vertex_id],
        gear: Gear[T_vertex, T_vertex_id, T_weight, T_labels],
        next_edges: Optional[
            BNextWeightedEdges[
                T_vertex,
                BSearchShortestPathFlex[T_vertex, T_vertex_id, T_weight, T_labels],
                T_weight,
            ],
        ] = None,
        *,
        next_labeled_edges: Optional[
            BNextWeightedLabeledEdges[
                T_vertex,
                BSearchShortestPathFlex[T_vertex, T_vertex_id, T_weight, T_labels],
                T_weight,
                T_labels,
            ],
        ] = None,
    ) -> None:
        self._vertex_to_id = vertex_to_id
        self._gear = gear

        (
            self._next_edges,
            self._labeled_edges,
        ) = _create_unified_next_weighted_bidirectional(next_edges, next_labeled_edges)

    def start_from(
        self,
        start_and_goal_vertex: Optional[tuple[T_vertex, T_vertex]] = None,
        *,
        start_and_goal_vertices: Optional[
            tuple[Iterable[T_vertex], Iterable[T_vertex]]
        ] = None,
        build_path: bool = False,
        calculation_limit: Optional[int] = None,
        fail_silently: bool = False,
    ) -> tuple[T_weight, Path[T_vertex, T_vertex_id, T_labels]]:
        """
        Start the search both from a start vertex and a goal vertex, resp. both
        from a set of start vertices and a set of goal vertices. Return the
        length of a shortest (sum of edge weights) path between the/a start vertex
        and the/a goal vertex. If building a path was requested, also return the path,
        and otherwise, return a dummy path object.

        If the search ends without having found a path, raise KeyError,
        or, if a silent fail is demanded, return infinity and a dummy path object.
        Here, infinity means the value for infinite distance that is defined by the
        used `Gear` (as provided by a call of gear.infinity()).

        NoGraphs gives no guarantees about a dummy path. Do not use it for anything.

        :param start_and_goal_vertex: The start vertex and the goal vertex of the
            search. If None, provide start_and_goal_vertices.

        :param start_and_goal_vertices: The set of start vertices and the set of
            goal vertices of the search. Only allowed if start_and_goal_vertex
            equals None.

        :param build_path: If true, build and return a path of the minimum possible
            length.

        :param calculation_limit: If provided, maximal number of vertices to process
            (read in) from your graph in each of the searches in one of the two
            directions. If it is exceeded, a RuntimeError will be raised.

        :param fail_silently: If no path can be found, fail silently (see above)
            instead of raising an exception.
        """
        # steps that other strategies do in start_from

        _search_needs_search_object(self, BSearchShortestPathFlex)

        start_vertices: Iterable[T_vertex]
        goal_vertices: Iterable[T_vertex]
        if start_and_goal_vertex is not None:
            if start_and_goal_vertices is not None:
                raise RuntimeError(
                    "Both start_and_goal_vertex and "
                    + "start_and_goal_vertices provided."
                )
            start_vertex, goal_vertex = start_and_goal_vertex
            start_vertices, goal_vertices = [start_vertex], [goal_vertex]
        else:
            if start_and_goal_vertices is None:
                raise RuntimeError(
                    "Neither start_and_goal_vertex nor "
                    + "start_and_goal_vertices provided."
                )
            start_vertices, goal_vertices = start_and_goal_vertices

        paths_forwards, predecessors_forwards, attributes_forwards = create_paths(
            build_path,
            self._gear,
            self._labeled_edges,
            self._vertex_to_id,
            start_vertices,
        )
        paths_backwards, predecessors_backwards, attributes_backwards = create_paths(
            build_path,
            self._gear,
            self._labeled_edges,
            self._vertex_to_id,
            goal_vertices,
        )

        # Copy Gear attributes into method scope (faster access)
        infinity = self._gear.infinity()

        # Copy Traversal attributes into method scope (faster access)
        labeled_edges = self._labeled_edges
        maybe_vertex_to_id = (
            None if self._vertex_to_id == vertex_as_id else self._vertex_to_id
        )  # Case vertex_as_id: not apply; T_vertex_id > T_vertex
        zero = self._gear.zero()

        # Prepare limit check done by zero check
        if calculation_limit is None:
            calculation_limit_forwards = calculation_limit_backwards = -1  # unused
        else:
            calculation_limit += 1
            calculation_limit_forwards = calculation_limit_backwards = calculation_limit

        # Copy traversal specific attributes into method scope
        next_edges_forwards, next_edges_backwards = self._next_edges

        # Get references of used gear objects and methods (avoid attribute resolution)
        (
            _,
            predecessors_sequence_forwards,
            predecessors_wrapper_forwards,
        ) = access_to_vertex_mapping_expect_none(predecessors_forwards)
        (
            _,
            predecessors_sequence_backwards,
            predecessors_wrapper_backwards,
        ) = access_to_vertex_mapping_expect_none(predecessors_backwards)

        (
            _,
            attributes_sequence_forwards,
            attributes_wrapper_forwards,
        ) = access_to_vertex_mapping_expect_none(attributes_forwards)
        (
            _,
            attributes_sequence_backwards,
            attributes_wrapper_backwards,
        ) = access_to_vertex_mapping_expect_none(attributes_backwards)

        # ----- Initialize method specific bookkeeping -----

        # Unique number, that prevents heapq from sorting by vertices in case of a
        # tie in the sort field, because vertices do not need to be pairwise
        # comparable. The integers from -5 to 256 are used first, because they are
        # internalized (pre-calculated, and thus fastest). We count downwards like we
        # do in A* search. There, it is preferable, because a LIFO behavior makes A*
        # often faster. Here, we do it simply to do it the same way.
        # (We use two counters so that "the same integers" can be used in the search
        #  in both directions)
        unique_no_forwards = itertools.count(256, -1)
        unique_no_backwards = itertools.count(256, -1)

        # Explicitly list start vertices and their id. Needed several times.
        start_vertices_and_ids_forwards = tuple(
            iter_start_vertices_and_ids(start_vertices, self._vertex_to_id)
        )
        start_vertices_and_ids_backwards = tuple(
            iter_start_vertices_and_ids(goal_vertices, self._vertex_to_id)
        )

        # Get the right class for storing a path (labeled or not)
        path_cls: type[Path]
        if labeled_edges:
            path_cls = PathOfLabeledEdges[T_vertex, T_vertex_id, T_labels]
        else:
            path_cls = PathOfUnlabeledEdges[T_vertex, T_vertex_id, T_labels]

        # At least one start vertex and one goal vertex is necessary to find
        # a result. The case of having no start or goal vertex is handled
        # manually here, because if there is no start vertex, the step
        # "choose an arbitrary start vertex..." will fail for this search strategy.
        if (
            len(start_vertices_and_ids_forwards) == 0
            or len(start_vertices_and_ids_backwards) == 0
        ):
            return self._search_failed(path_cls, infinity, fail_silently)

        # Detect if a start vertex is also goal vertex, and report result manually.
        # (Without this manual handling, a non-self loop from such a start vertex back
        #  to itself with a length > 0 would be reported as smallest distance. This
        #  would be unexpected for users, since they expect that the distance from a
        #  vertex to itself is always 0.)
        common_vertex_ids = set(
            v_id for v, v_id in start_vertices_and_ids_forwards
        ).intersection(v_id for v, v_id in start_vertices_and_ids_backwards)
        if common_vertex_ids:
            for c_vertex, cv_id in start_vertices_and_ids_forwards:
                if cv_id in common_vertex_ids:
                    p = path_cls.from_vertex(c_vertex)
                    return zero, p

        # At start, most of the distances from a vertex to a start vertex are not
        # known. If accessed for comparison for possibly better distances, infinity
        # is used, if no other value is given. Each start vertex has distance 0
        # from a start vertex (itself), if not defined otherwise.
        distances_forwards = define_distances(
            self._gear,
            None,
            (
                (vertex_id, zero)
                for vertex, vertex_id in start_vertices_and_ids_forwards
            ),
            False,
        )
        distances_backwards = define_distances(
            self._gear,
            None,
            (
                (vertex_id, zero)
                for vertex, vertex_id in start_vertices_and_ids_backwards
            ),
            False,
        )

        # Get references of used gear objects and methods (avoid attribute resolution)
        (
            _,
            distances_sequence_forwards,
            distances_wrapper_forwards,
        ) = access_to_vertex_mapping(distances_forwards)
        (
            _,
            distances_sequence_backwards,
            distances_wrapper_backwards,
        ) = access_to_vertex_mapping(distances_backwards)

        # So far, the start vertices are to be visited.
        # (No index exception possible at the following indexed access)
        # The following lists are used as collection.heapq of tuples, the lowest
        # distance first.
        to_visit_forwards = [
            (distances_sequence_forwards[vertex_id], next(unique_no_forwards), vertex)
            for vertex, vertex_id in start_vertices_and_ids_forwards
        ]
        heapify(to_visit_forwards)
        to_visit_backwards = [
            (distances_sequence_backwards[vertex_id], next(unique_no_backwards), vertex)
            for vertex, vertex_id in start_vertices_and_ids_backwards
        ]
        heapify(to_visit_backwards)

        # No connection between forwards and backwards search found so far
        # (The value of best_connection_node is not used. So, we initialize it just to
        #  any vertex)
        best_path_length = infinity
        best_connecting_node = start_vertices_and_ids_forwards[0][0]
        others_path_weight = zero

        # ----- Loop -----
        for (
            to_visit,
            distances_sequence,
            distances_wrapper,
            distances_sequence_other,
            distances_wrapper_other,
            next_edges,
            paths,
            predecessors_sequence,
            predecessors_wrapper,
            attributes,
            attributes_sequence,
            attributes_wrapper,
            unique_no,
            forwards,
        ) in itertools.cycle(
            (
                (
                    to_visit_forwards,
                    distances_sequence_forwards,
                    distances_wrapper_forwards,
                    distances_sequence_backwards,
                    distances_wrapper_backwards,
                    next_edges_forwards,
                    paths_forwards,
                    predecessors_sequence_forwards,
                    predecessors_wrapper_forwards,
                    attributes_forwards,
                    attributes_sequence_forwards,
                    attributes_wrapper_forwards,
                    unique_no_forwards,
                    True,
                ),
                (
                    to_visit_backwards,
                    distances_sequence_backwards,
                    distances_wrapper_backwards,
                    distances_sequence_forwards,
                    distances_wrapper_forwards,
                    next_edges_backwards,
                    paths_backwards,
                    predecessors_sequence_backwards,
                    predecessors_wrapper_backwards,
                    attributes_backwards,
                    attributes_sequence_backwards,
                    attributes_wrapper_backwards,
                    unique_no_backwards,
                    False,
                ),
            )
        ):
            if not to_visit:
                # If the same edges are reported in both directions and there is a
                # path between a start vertex and a goal vertex, then neither the
                # forward nor the backward iteration can become exhausted before a
                # solution is found. Thus, running out of vertices in one of the
                # iterations mean there is no solution of the adjacency functions have
                # errors.
                return self._search_failed(path_cls, infinity, fail_silently)

            # Visit path with the lowest distance first
            path_weight, _, vertex = heappop(to_visit)

            # A vertex can get added to the heap multiple times. We want to process
            # it only once, the first time it is removed from the heap, because this
            # is the case with the shortest distance from start.
            v_id: T_vertex_id = (
                maybe_vertex_to_id(vertex)  # type: ignore[assignment]
                if maybe_vertex_to_id
                else vertex
            )

            if distances_sequence[v_id] < path_weight:
                continue

            if infinity != best_path_length <= path_weight + others_path_weight:
                # (Infinity handled manually: if best_path_length is infinity,
                # the sum to the right has to be regarded as being smaller, disregarding
                # its value, that could be higher in case of an overflow over
                # infinity.)
                break  # Finished. Nothing shorter than best_path_length will come.

            # store our length for the check when expanding in the opposite directions
            others_path_weight = path_weight

            if calculation_limit and not (
                (calculation_limit_forwards := calculation_limit_forwards - 1)
                if forwards
                else (calculation_limit_backwards := calculation_limit_backwards - 1)
            ):
                raise RuntimeError("Number of visited vertices reached limit")

            # We will expand this vertex

            # Expand vertex. New neighbors are one edge count deeper than vertex.
            for edge in next_edges(vertex, self):
                neighbor, weight = edge[0], edge[1]

                n_path_weight = weight + path_weight
                # (Distance values equal to or higher than the chosen infinity value of
                # the gear are invalid and cannot be handled further.)
                if infinity <= n_path_weight:
                    self._gear.raise_distance_infinity_overflow_error(n_path_weight)

                # If the found path to the neighbor is not shorter than the shortest
                # such path found so far, we can safely ignore it. Otherwise, it is a
                # new candidate for a shortest path to the neighbor, and we push it to
                # the heap.
                n_id: T_vertex_id = (
                    maybe_vertex_to_id(neighbor)  # type: ignore[assignment]
                    if maybe_vertex_to_id
                    else neighbor
                )

                try:
                    if distances_sequence[n_id] <= n_path_weight:
                        continue
                    distances_sequence[n_id] = n_path_weight
                except IndexError:
                    # n_id not in distances_collection. To be regarded as value
                    # infinity, i.e., n_path_weight is smaller.
                    distances_wrapper.extend_and_set(n_id, n_path_weight)

                # If we are to generate a path, we have to do it here, since the
                # edge we have to add to the path prefix is not stored on the heap
                if build_path:
                    try:
                        predecessors_sequence[n_id] = vertex
                    except IndexError:
                        predecessors_wrapper.extend_and_set(n_id, vertex)
                    if labeled_edges:
                        # self._labeled_edges  -> next_edges (a NextWeightedEdges) is a
                        # NextWeightedLabeledEdges -> edge[-1] is a T_labels.
                        data_of_edge: T_labels = edge[-1]  # type: ignore[assignment]
                        try:
                            attributes_sequence[n_id] = data_of_edge
                        except IndexError:
                            attributes_wrapper.extend_and_set(n_id, data_of_edge)

                heappush(to_visit, (n_path_weight, next(unique_no), neighbor))

                # If we already have an estimation for the distance from the neighbor
                # to the other end of the search and the resulting total distance
                # is better than the best we have seen so far, store the length and
                # the neighbor vertex.
                try:
                    if (
                        n_path_weight + distances_sequence_other[n_id]
                        >= best_path_length
                    ):
                        continue
                    best_path_length = n_path_weight + distances_sequence_other[n_id]
                    best_connecting_node = neighbor
                except IndexError:
                    # n_id not in distances_collection. To be regarded as value
                    # infinity, i.e., best_path_length is equal or smaller.
                    pass

            path = (
                path_cls.from_bidirectional_search(
                    paths_forwards, paths_backwards, best_connecting_node
                )
                if build_path
                else path_cls.of_nothing()
            )
        return best_path_length, path

    @staticmethod
    def _search_failed(
        path_cls: type[Path], infinity: T_weight, fail_silently: bool
    ) -> tuple[T_weight, Path[T_vertex, T_vertex_id, T_labels]]:
        """For a silent fail, return the value that marks that the search has failed:
        A tuple of the infinity value of the weights and an empty path. If no
        silent fail is requested, raise a key error telling that no goal vertex
        has been found."""
        if fail_silently:
            return infinity, path_cls.of_nothing()
        else:
            raise KeyError("No path to (a) goal vertex found")


class BSearchShortestPath(
    Generic[T_vertex, T_weight, T_labels],
    BSearchShortestPathFlex[T_vertex, T_vertex, Union[T_weight, float], T_labels],
):
    """
    Eases the use of `BSearchShortestPathFlex` for typical cases.
    For documentation of functionality and parameters, see there.

    .. code-block:: python

       BSearchShortestPath[T_vertex, T_weight, T_labels](*args, **keywords)

    is a short form for

    .. code-block:: python

       BSearchShortestPathFlex[
           T_vertex, T_vertex, Union[T_weight, float], T_labels],
      ](nog.vertex_as_id, nog.GearDefault(), *args, **keywords)

    Implication:

    - `GearDefault` is used, see there how it and its superclass work
    - The used weights are defined by Union[T_weight, float], see `GearDefault`
    - T_vertex is bound to Hashable (T_vertex is used as `T_vertex_id`, see there)
    """

    def __init__(
        self,
        next_edges: Optional[
            BNextWeightedEdges[
                T_vertex,
                BSearchShortestPathFlex[
                    T_vertex, T_vertex, Union[T_weight, float], T_labels
                ],
                T_weight,
            ],
        ] = None,
        *,
        next_labeled_edges: Optional[
            BNextWeightedLabeledEdges[
                T_vertex,
                BSearchShortestPathFlex[
                    T_vertex, T_vertex, Union[T_weight, float], T_labels
                ],
                T_weight,
                T_labels,
            ],
        ] = None,
    ) -> None:
        super().__init__(
            vertex_as_id,
            GearDefault(),
            next_edges,
            next_labeled_edges=next_labeled_edges,
        )



from __future__ import annotations

from abc import ABC

from collections.abc import (
    Callable,
    Iterator,
    Iterable,
    MutableSet,
    MutableMapping,
)
from typing import TypeVar, Optional, Any, Union, cast, Generic

from ._gear_collections import (
    get_wrapper_from_vertex_set,
    get_wrapper_from_vertex_mapping,
)
from ._gears import (
    GearWithoutDistances,
    Gear,
    VertexIdSet,
    VertexIdToVertexMapping,
    VertexIdToDistanceMapping,
    VertexIdToPathEdgeDataMapping,
)
from ._paths import (
    Paths,
    PathsOfUnlabeledEdges,
    PathsOfLabeledEdges,
    DummyPredecessorOrAttributesMapping,
    PathsDummy,
)
from ._types import (
    T_vertex,
    T_vertex_id,
    T_weight,
    T_labels,
    VertexToID,
    vertex_as_id,
    WeightedUnlabeledOutEdge,
    WeightedLabeledOutEdge,
    UnweightedLabeledOutEdge,
    LabeledOutEdge,
    OutEdge,
)


# --------------- classes -------------


class StrRepr:
    """Provides a specifically "normalized" string representation of data."""

    def __init__(self, s: str) -> None:
        self.s = s

    @classmethod
    def from_iterable(cls, i: Iterable[tuple[Any, Any]]) -> StrRepr:
        """
        Provides a string representation of an iterable of key/value tuples,
        that look like the output from a dict with these items.

        (The 'keys' do not need to be hashable.)
        """
        return cls("{" + ", ".join(repr(k) + ": " + repr(v) for k, v in i) + "}")

    @classmethod
    def from_set(cls, c: MutableSet[Any]) -> StrRepr:
        """
        Provides a string representation of a *MutableSet*,
        that looks like the string representation of a *Set* with these items,
        but the elements are lexicographically sorted.

        The result is independent of the methods repr() and str() of the
        *MutableSet*. (The elements do not need to be hashable.)
        """
        return cls("{" + ", ".join(sorted(repr(k) for k in c)) + "}")

    def __repr__(self) -> str:
        return self.s


class Strategy(ABC, Generic[T_vertex, T_vertex_id, T_labels]):
    """Base class of the traversal strategies and search strategies of NoGraphs."""

    def _improve_state(
        self, state: dict[str, Any], vertices: Optional[Iterable[T_vertex]] = None
    ) -> None:
        """Improve the state description

        :param state: State in current form
        :param vertices: If the strategy can provide additional state data w.r.t. these
            vertices, it will do so.
        """
        pass

    def state_to_str(self, vertices: Optional[Iterable[T_vertex]] = None) -> str:
        """Return a human-readable description of the public state of the strategy as
        a string.

        Implementation details, not covered by the semantic versioning:

        Currently, the method aims at providing a uniform behaviour over different
        platforms (*CPython* and *PyPy*) and collection types (Gears with different
        *MutableSet* and *MutableMapping* implementations). It behaves roughly as
        follows:

        - A *MutableSet*, e.g. attribute *visited*, is described similar to a *set*,
          but items are sorted lexicographically in their string representations
          (this bridges differences between *CPython* and *PyPy*).

        - Attribute *Paths* is described similar to a *dict* (although keys can contain
          unhashable values, and only paths for the given *vertices* are described).

        - A *MutableMapping*, e.g. attribute *distance*, is described similarly to a
          *dict*, also in cases, where it is not a *dict*, and although the items
          for only the given *vertices* are described.

        :param vertices: If the strategy can provide additional state data w.r.t. these
            vertices, it will do so.
        """
        state = dict((k, v) for k, v in self.__dict__.items() if k[0] != "_")
        self._improve_state(state, vertices)
        return str(state)


# --------------- exported types -------------

# todo: Warning: The following types are manually documented in api.rst

# next vertices and next edges functions for traversals
# that work with and without weights

T_strategy = TypeVar("T_strategy", bound=Strategy)

NextVertices = Callable[[T_vertex, T_strategy], Iterable[T_vertex]]

NextEdges = Callable[[T_vertex, T_strategy], Iterable[OutEdge[T_vertex, Any, Any]]]

NextLabeledEdges = Callable[
    [T_vertex, T_strategy], Iterable[LabeledOutEdge[T_vertex, Any, T_labels]]
]

# next edges functions for traversal that work with weights
NextWeightedEdges = Callable[
    [T_vertex, T_strategy],
    Iterable[
        Union[
            WeightedUnlabeledOutEdge[T_vertex, T_weight],
            WeightedLabeledOutEdge[T_vertex, T_weight, Any],
        ]
    ],
]

NextWeightedLabeledEdges = Callable[
    [T_vertex, T_strategy],
    Iterable[WeightedLabeledOutEdge[T_vertex, T_weight, T_labels]],
]

# The same, but as a tuple, for bidirectional search strategies

BNextVertices = tuple[
    NextVertices[T_vertex, T_strategy],
    NextVertices[T_vertex, T_strategy],
]

BNextEdges = tuple[
    NextEdges[T_vertex, T_strategy],
    NextEdges[T_vertex, T_strategy],
]

BNextLabeledEdges = tuple[
    NextLabeledEdges[T_vertex, T_strategy, T_labels],
    NextLabeledEdges[T_vertex, T_strategy, T_labels],
]

BNextWeightedEdges = tuple[
    NextWeightedEdges[T_vertex, T_strategy, T_weight],
    NextWeightedEdges[T_vertex, T_strategy, T_weight],
]

BNextWeightedLabeledEdges = tuple[
    NextWeightedLabeledEdges[T_vertex, T_strategy, T_weight, T_labels],
    NextWeightedLabeledEdges[T_vertex, T_strategy, T_weight, T_labels],
]


# --------------- package internal types -------------

NextEdgesOrVertices = Callable[
    [T_vertex, T_strategy],
    Iterable[
        Union[
            T_vertex,
            WeightedUnlabeledOutEdge[T_vertex, Any],
            UnweightedLabeledOutEdge[T_vertex, T_labels],
            WeightedLabeledOutEdge[T_vertex, Any, T_labels],
        ]
    ],
]

NextWeightedMaybeLabeledEdges = Callable[
    [T_vertex, T_strategy],
    Iterable[
        Union[
            WeightedUnlabeledOutEdge[T_vertex, T_weight],
            WeightedLabeledOutEdge[T_vertex, T_weight, T_labels],
        ]
    ],
]


# The same, but as a tuple, for bidirectional search strategies

BNextEdgesOrVertices = tuple[
    NextEdgesOrVertices[T_vertex, T_strategy, T_labels],
    NextEdgesOrVertices[T_vertex, T_strategy, T_labels],
]

BNextWeightedMaybeLabeledEdges = tuple[
    NextWeightedMaybeLabeledEdges[T_vertex, T_strategy, T_weight, T_labels],
    NextWeightedMaybeLabeledEdges[T_vertex, T_strategy, T_weight, T_labels],
]


# --------------- internal support functions -------------


def iter_start_ids(
    start_vertices: Iterable[T_vertex], vertex_to_id: VertexToID[T_vertex, T_vertex_id]
) -> Iterable[T_vertex_id]:
    """Compute vertex ids for given start vertices and allow for iterating
    them"""
    if vertex_to_id == vertex_as_id:
        # If the identity function (in a mathematical sense)
        # vertex_as_id is used with correct typing, this means that
        # T_vertex is a subtype of T_vertex_id (typically: identical).
        # So, instead of applying the function, we could just cast the vertices
        # to vertex ids. For improved performance, we cast the whole iterator.
        return cast(Iterable[T_vertex_id], start_vertices)

    return (vertex_to_id(vertex) for vertex in start_vertices)


def iter_start_vertices_and_ids(
    start_vertices: Iterable[T_vertex], vertex_to_id: VertexToID[T_vertex, T_vertex_id]
) -> Iterable[tuple[T_vertex, T_vertex_id]]:
    """Compute vertex ids for given start vertices and allow for iterating
    pairs of vertex and vertex id."""
    if vertex_to_id == vertex_as_id:
        # If the identity function (in a mathematical sense)
        # vertex_as_id is used with correct typing, this means that
        # T_vertex is a subtype of T_vertex_id (typically: identical).
        # So, instead of applying the function, we could just cast the vertices
        # to vertex ids. For improved performance, we cast the whole iterator.
        vertices_and_vertices = ((vertex, vertex) for vertex in start_vertices)
        vertices_and_ids = cast(
            Iterator[tuple[T_vertex, T_vertex_id]], vertices_and_vertices
        )
        return vertices_and_ids

    return ((vertex, vertex_to_id(vertex)) for vertex in start_vertices)


def define_visited(
    gear: GearWithoutDistances[T_vertex, T_vertex_id, T_labels],
    already_visited: Optional[VertexIdSet[T_vertex_id]],
    iter_start_ids: Iterable[T_vertex_id],
    is_tree: bool,
) -> VertexIdSet[T_vertex_id]:
    """Use and return already_visited, if provided, for storing visited vertices,
    and otherwise a new VertexIdSet. Mark start vertices as visited."""
    if already_visited is None:
        return gear.vertex_id_set(() if is_tree else iter_start_ids)

    if not is_tree:
        if (wrapper := get_wrapper_from_vertex_set(already_visited)) is None:
            method_add = already_visited.add
            for v_id in iter_start_ids:
                method_add(v_id)
        else:
            wrapper.update_from_keys(iter_start_ids)
    return already_visited


def define_distances(
    gear: Gear[T_vertex, T_vertex_id, T_weight, T_labels],
    known_distances: Optional[VertexIdToDistanceMapping[T_vertex_id, T_weight]],
    iter_start_ids_and_distances: Iterable[tuple[T_vertex_id, T_weight]],
    is_tree: bool,
) -> VertexIdToDistanceMapping[T_vertex_id, T_weight]:
    """Use and return known_distances, if provided, for storing vertex distances, and
    otherwise new VertexIdToDistanceMapping. Store the distances given for the
    start vertices."""
    if known_distances is None:
        return gear.vertex_id_to_distance_mapping(iter_start_ids_and_distances)

    if not is_tree:
        if (wrapper := get_wrapper_from_vertex_mapping(known_distances)) is None:
            method_setdefault = known_distances.setdefault
            for v_id, distance in iter_start_ids_and_distances:
                method_setdefault(v_id, distance)
        else:
            wrapper.update_default(iter_start_ids_and_distances)

    return known_distances


def create_paths(
    build_paths: bool,
    gear: GearWithoutDistances[T_vertex, T_vertex_id, T_labels],
    labeled_edges: bool,
    vertex_to_id: VertexToID[T_vertex, T_vertex_id],
    start_vertices: Iterable[T_vertex],
) -> tuple[
    Paths[T_vertex, T_vertex_id, T_labels],
    VertexIdToVertexMapping[T_vertex_id, T_vertex],
    VertexIdToPathEdgeDataMapping[T_vertex_id, T_labels],
]:
    """Translate from configuration of path generation to setting of
    paths, predecessors and attributes collection. Store empty paths for start
    vertices. If no paths should be build, create dummy Paths container."""

    if build_paths:
        # Create container for predecessors.
        # From each start vertex, store an empty paths to itself.
        predecessor = gear.vertex_id_to_vertex_mapping(
            (
                (vertex_id, vertex)
                for vertex, vertex_id in iter_start_vertices_and_ids(
                    start_vertices, vertex_to_id
                )
            )
        )
        paths: Paths[T_vertex, T_vertex_id, T_labels]
        attributes: Optional[VertexIdToPathEdgeDataMapping[T_vertex_id, T_labels]]
        if labeled_edges:
            attributes = gear.vertex_id_to_path_attributes_mapping(())
            paths = PathsOfLabeledEdges[T_vertex, T_vertex_id, T_labels](
                predecessor, attributes, vertex_to_id
            )
        else:
            paths = PathsOfUnlabeledEdges[T_vertex, T_vertex_id](
                predecessor, vertex_to_id
            )
            attributes = DummyPredecessorOrAttributesMapping[T_vertex_id, T_labels]()
        return paths, predecessor, attributes
    else:
        return (
            PathsDummy[T_vertex, T_vertex_id, T_labels](vertex_to_id),
            DummyPredecessorOrAttributesMapping[T_vertex_id, T_vertex](),
            DummyPredecessorOrAttributesMapping[T_vertex_id, T_labels](),
        )


class NoIterator(Generic[T_vertex]):
    """An iterator that raises RuntimeError instead of iterating.

    When the application requests an iterator from a traversal that has not been
    started so far and uses this iterator, this is a programming error in the
    application: When the traversal is started, the iterator will be replaced, but the
    application will still have and maybe use the one it has requested to early. Thus,
    the traversal returns a _NoIterator in this case, so that any use of it clearly
    shows the problem.
    """

    def __next__(self) -> T_vertex:
        """
        >>> next(NoIterator())
        Traceback (most recent call last):
        RuntimeError: Traversal not started, iteration not possible
        """
        raise RuntimeError("Traversal not started, iteration not possible")

    def __iter__(self) -> Iterator[T_vertex]:
        """
        >>> iter(NoIterator())
        Traceback (most recent call last):
        RuntimeError: Traversal not started, iteration not possible
        """
        raise RuntimeError("Traversal not started, iteration not possible")


class NoVisitedSet(MutableSet[T_vertex_id]):
    """A MutableSet for vertex ids that raises an exception on each operation.

    When the application accesses a state attribute of a traversal and the attribute is
    not initialized so far (i.e., before the first vertex has been expanded or
    reported), the traversal shows this problem to the application by returning
    a clearly functionless _NoVisitedSet in attribute *visited*.
    """

    def __contains__(self, key: object) -> bool:
        raise RuntimeError("Traversal not started, no data to be accessed")

    def __iter__(self) -> Iterator[T_vertex_id]:
        raise RuntimeError("Traversal not started, no data to be accessed")

    def __len__(self) -> int:
        raise RuntimeError("Traversal not started, no data to be accessed")

    def discard(self, value: T_vertex_id) -> None:
        raise RuntimeError("Traversal not started, no data to be accessed")

    def add(self, value: T_vertex_id) -> None:
        raise RuntimeError("Traversal not started, no data to be accessed")


class NoDistancesMapping(MutableMapping[T_vertex_id, T_weight]):
    """A MutableMapping from vertex ids to distances that raises an exception on each
    operation.

    When the application accesses a state attribute of a traversal and the attribute is
    not initialized so far (i.e., before the first vertex has been expanded or
    reported), the traversal shows this problem to the application by returning
    a clearly functionless _NoDistancesMapping in attribute *distances*.
    """

    def __getitem__(self, key: T_vertex_id) -> T_weight:
        raise RuntimeError("Traversal not started, no data to be accessed")

    def __delitem__(self, key: T_vertex_id) -> None:
        raise RuntimeError("Traversal not started, no data to be accessed")

    def __iter__(self) -> Iterator[T_vertex_id]:
        raise RuntimeError("Traversal not started, no data to be accessed")

    def __len__(self) -> int:
        raise RuntimeError("Traversal not started, no data to be accessed")

    def __contains__(self, key: object) -> bool:
        raise RuntimeError("Traversal not started, no data to be accessed")

    def __setitem__(self, key: T_vertex_id, value: T_weight) -> None:
        raise RuntimeError("Traversal not started, no data to be accessed")



import nographs as nog  # noqa: F401 (used in doctests, undetected by flake 8)

# noinspection PyProtectedMember
import nographs._gear_collections as gear_collections  # noqa: F401 (used by doctests)


class ProtocolAndABCNotImplementedErrors:
    """-- Abstract methods of protocols and ABCs.

    If the application calls them and ignores that they are abstract, an assertion
    is to be raised to inform the application about its mistake.
    Check, if this mechanism is correctly implemented.

    Note: The following calls are all illegal w.r.t. typing (only the number of
    parameters is correct): Instance methods are called like a classmethod would,
    the given argument for parameter self has the wrong type, and other arguments may
    be illegal, too, and the generic parameters are missing. But all this does not
    matter here, since the methods are to raise NotImplementedError directly and in
    all cases.

    >>> nog.GettableSettableForGearProto.__getitem__(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.GettableSettableForGearProto.__setitem__(None, None, None)
    Traceback (most recent call last):
    NotImplementedError


    >>> nog.SequenceForGearProto.__len__(None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.SequenceForGearProto.append(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.SequenceForGearProto.extend(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.SequenceForGearProto.__iter__(None)
    Traceback (most recent call last):
    NotImplementedError


    >>> nog.VertexSequenceWrapperForSetProto.sequence(None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSequenceWrapperForSetProto.default(None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSequenceWrapperForSetProto.extend_and_set(None, None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSequenceWrapperForSetProto.update_from_keys(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSequenceWrapperForSetProto.index_and_bit_method(None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSequenceWrapperForSetProto._from_iterable(None, None)
    Traceback (most recent call last):
    NotImplementedError


    >>> nog.VertexSequenceWrapperForMappingProto.sequence(None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSequenceWrapperForMappingProto.default(None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSequenceWrapperForMappingProto.extend_and_set(None, None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSequenceWrapperForMappingProto.update_from_keys_values(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSequenceWrapperForMappingProto.update_default(None, None)
    Traceback (most recent call last):
    NotImplementedError


    >>> nog.VertexSetWrappingSequence.__iter__(None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSetWrappingSequence.__len__(None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSetWrappingSequence.update_from_keys(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSetWrappingSequence._from_iterable(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.VertexSetWrappingSequence.index_and_bit_method(None)
    Traceback (most recent call last):
    NotImplementedError
    """


class CallToPrivateDummyFunctionality:
    """
    The following classes are library private. Only NoGraphs is expected the
    see and use them. And NoGraphs is expected to never call their methods. Thus,
    if a call ever happens, this would be unexpected and probly due to an error
    in NoGraphs. An assertion error is to be raised to signal this.

    >>> _GettableSettableForGearAssertNoCall = (
    ...     # noinspection PyProtectedMember
    ...     gear_collections._GettableSettableForGearAssertNoCall
    ... )

    >>> _GettableSettableForGearAssertNoCall.__getitem__(None, None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen

    >>> _GettableSettableForGearAssertNoCall.__setitem__(None, None, None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen

    >>> # noinspection PyProtectedMember
    >>> _VertexSequenceWrapperAssertNoCall = (
    ...     # noinspection PyProtectedMember
    ...     gear_collections._VertexSequenceWrapperAssertNoCall
    ... )

    >>> _VertexSequenceWrapperAssertNoCall.sequence(None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen

    >>> _VertexSequenceWrapperAssertNoCall.default(None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen

    >>> _VertexSequenceWrapperAssertNoCall.extend_and_set(None, None, None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen

    >>> _VertexSequenceWrapperAssertNoCall.update_from_keys(None, None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen

    >>> _VertexSequenceWrapperAssertNoCall.index_and_bit_method(None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen

    >>> _VertexSequenceWrapperAssertNoCall.update_from_keys_values(None, None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen

    >>> _VertexSequenceWrapperAssertNoCall.update_default(None, None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen

    >>> # noinspection PyProtectedMember
    >>> _VertexSequenceWrapperAssertNoCall._from_iterable(None, None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen
    """


# --------- edge cases of the functionality -----------


class GearCollectionFunctionalityMainlyOnlyForAppCode:
    """
    Test the functionality, that NoGraphs inlines, does not use itself,
    and that is mainly there for application code only.
    Exception: see documentation of VertexSequenceWrapperForSetAndMapping.

    >>> list_factory = lambda: list()
    >>> ws = nog.VertexSetWrappingSequenceNoBitPacking(
    ...     list_factory, 1024, [1, 3])
    >>> ws
    {1, 3}
    >>> len(ws.sequence())
    1026
    >>> "a" in ws  # Case "is not instance(key, int)"
    False
    >>> 1026 in ws  # Case "IndexError"
    False
    >>> 1 in ws  # Case "return True"
    True
    >>> 2 in ws  # Case "return False"
    False
    >>> len(ws)
    2
    >>> ws.add(2)
    >>> ws
    {1, 2, 3}
    >>> ws.discard(3)
    >>> ws
    {1, 2}
    >>> ws.discard(3)
    >>> ws.add(1026)  # Case "IndexError"
    >>> ws
    {1, 2, 1026}
    >>> len(ws.sequence())
    2051
    >>> ws.discard(2051)   # Case "IndexError"
    >>> ws | ({4})  # Calls _from_iterable(iterable) to create new set
    {1, 2, 4, 1026}

    >>> list_factory = lambda: list()
    >>> ws = nog.VertexSetWrappingSequenceBitPacking(
    ...     list_factory, 128, [1, 3])
    >>> ws
    {1, 3}
    >>> len(ws.sequence())
    129
    >>> "a" in ws  # Case "is not instance(key, int)"
    False
    >>> 129*8 in ws  # Case "IndexError"
    False
    >>> 1 in ws  # Case "return True"
    True
    >>> 2 in ws  # Case "return False"
    False
    >>> len(ws)
    2
    >>> ws.add(2)
    >>> ws
    {1, 2, 3}
    >>> ws.discard(3)
    >>> ws
    {1, 2}
    >>> ws.discard(3)
    >>> ws.add(129*8)  # Case "IndexError"
    >>> ws
    {1, 2, 1032}
    >>> len(ws.sequence())
    258
    >>> ws.discard(258*8)   # Case "IndexError"
    >>> ws | ({4})  # Calls _from_iterable(iterable) to create new set
    {1, 2, 4, 1032}

    >>> list_factory = lambda: list[float]()
    >>> ws = nog.VertexMappingWrappingSequence(
    ...     list_factory, float("infinity"), 1024, [(0, 0), (2, 2)])
    >>> ws.default()  # Gap marker / default value of the mapping emulation
    inf
    >>> ws.sequence()[:5]  # Given values are set, others are gap-marker / default
    [0, inf, 2, inf, inf]
    >>> len(ws.sequence())
    1025
    >>> ws
    {0: 0, 2: 2}
    >>> ws.update_default([(1, 1), (2, 2.5), (3, 3)])  # key 2 has value ->ignore change
    >>> ws
    {0: 0, 1: 1, 2: 2, 3: 3}
    >>> ws.update_default([(1025, 1025)])  # Gives IndexError -> extend_and_set
    >>> ws
    {0: 0, 1: 1, 2: 2, 3: 3, 1025: 1025}
    >>> ws[0]
    0
    >>> ws[4]  # Gives KeyError because gap marker
    Traceback (most recent call last):
    KeyError
    >>> ws[2048]  # Gives KeyError because outside
    Traceback (most recent call last):
    KeyError
    >>> "hallo" in ws  # str is outside key type int
    False
    >>> len(ws)
    5
    >>> del ws[1]
    >>> ws
    {0: 0, 2: 2, 3: 3, 1025: 1025}
    >>> len(ws.sequence())
    2050
    >>> ws[2050] = 2050
    >>> ws
    {0: 0, 2: 2, 3: 3, 1025: 1025, 2050: 2050}
    """


class GearCollectionTestRemainingCasesForUseByLibrary:
    """
    An example that uses VertexMappingWrappingSequenceWithNone
    (and calls methods of VertexMappingWrappingSequenceWithNone) with distances is
    already given in the docs.
    Missing, and given here, is an example that uses VertexSetWrappingSequence
    (calls methods of VertexSetByWrapper)
    and VertexMappingWrappingSequenceWithoutNone
    (calls methods of VertexMappingByWrapper with case without None).

    Fehlt: Aufruf von access_paths_to_vertex_mapping_expect_none, wobei unter dem
    Mapping keine VertexMappingByWrapperWithNone sondern eine VertexMappingByWrapper
    liegt. Ich dachte, predecessor von Paths nutzt ...expect_none, und wenn ein
    Array darunter liegt, wird von GearArrays... dann VertexMappingByWrapper und nicht
    VertexMappingByWrapperWithNone genutzt...

    >>> def next_vertices(i, _):
    ...     yield i+1
    >>> t = nog.TraversalDepthFirstFlex(
    ...     nog.vertex_as_id, nog.GearForIntVerticesAndIDsAndCFloats(), next_vertices
    ... )
    >>> t.paths.debug = True
    >>> v = t.start_from(0, build_paths=True, compute_depth = True).go_to(5)
    >>> t.paths[5]
    (0, 1, 2, 3, 4, 5)
    >>> t.depth
    5
    """


class GearCollectionTestsForDoNotCallCases:
    """
    A "real" set has no wrapper. So, the index_and_bit_method returned by
    nog.access_to_vertex_set is not intended to be called.
    >>> c = set()
    >>> res = gear_collections.access_to_vertex_set(c)
    >>> is_wrapper, gettable_settable, wrapper, uses_bits, index_and_bit_method = res
    >>> index_and_bit_method(None, None)
    Traceback (most recent call last):
    AssertionError: Call to a method of this object is not expected to ever happen
    """



from __future__ import annotations

import collections
import itertools
from collections.abc import Sequence, Mapping, Callable, Iterable
from typing import Union, Any, Literal, overload

from ._types import (
    T_vertex,
    T_weight,
    T_labels,
    OutEdge,
    WeightedOrLabeledFullEdge,
    AnyFullEdge,
)


# ---------- functions for handling test data given as some kind of edges -------------


@overload
def adapt_edge_index(
    index: Mapping[T_vertex, Iterable[T_vertex]],
    *,
    add_inverted: bool,
    attributes: Literal[False],
) -> Callable[[T_vertex, Any], Iterable[T_vertex]]:
    ...


@overload
def adapt_edge_index(
    index: Mapping[T_vertex, Iterable[OutEdge[T_vertex, T_weight, T_labels]]],
    *,
    add_inverted: bool,
    attributes: Literal[True],
) -> Callable[[T_vertex, Any], Iterable[OutEdge[T_vertex, T_weight, T_labels]]]:
    ...


@overload
def adapt_edge_index(
    index: Sequence[Iterable[T_vertex]],
    *,
    add_inverted: bool,
    attributes: Literal[False],
) -> Callable[[int, Any], Iterable[T_vertex]]:
    ...


@overload
def adapt_edge_index(
    index: Sequence[Iterable[OutEdge[int, T_weight, T_labels]]],
    *,
    add_inverted: bool,
    attributes: Literal[True],
) -> Callable[[int, Any], Iterable[OutEdge[int, T_weight, T_labels]]]:
    ...


def adapt_edge_index(
    index: Union[Mapping, Sequence], *, add_inverted: bool, attributes: bool
) -> Callable:
    """
    Read a graph from a Mapping (e.g. a Dict) or from a Sequence (e.g. a tuple
    or list, if integers are used as the vertices) and provide a neighbor function
    (`NextVertices` or `NextEdges`) from that data. Typically only used for test
    purposes.

    :param index: Mapping or Sequence with vertices as key, resp. as index. For each
        vertex, you provide:

        - either an Iterable of adjacent vertices (this defines edges
          without edge attributes)

        - or an Iterable of edges (with edge attributes) that start at this vertex.
          An edge to some neighbor (see `outgoing edge <outgoing_edges>`) needs
          to be a Sequence of the form
          (neighbor, labels) or (neighbor, weight) or (neighbor, weight, labels).

    :param attributes: Give True, if *index* provides labeled or weighted edges.
        In this case, you can use the resulting neighbor function as `NextEdges`
        function.

        Give False, if index provides neighbor vertices for a given vertex.
        In this case, you can use the resulting neighbor function as
        `NextVertices` function.

    :param add_inverted: If False, the resulting neighbor function will yield only the
        edges of *index*. If True, the function will also provide the
        same edges inverted (start and end vertex exchanged).

        Note: when this option is used, a
        copy of your graph will be held by the returned neighbor function.

    :return: Neighbor function that can be used as parameter for one of the traversal
        algorithms.
    """

    if add_inverted:
        if isinstance(index, Mapping):
            if attributes:
                return adapt_edge_iterable(
                    (
                        (v1,) + e_to
                        for v1, edges_to in index.items()
                        for e_to in edges_to
                    ),
                    add_inverted=add_inverted,
                    attributes=True,
                )
            else:
                return adapt_edge_iterable(
                    ((v1, e_to) for v1, edges_to in index.items() for e_to in edges_to),
                    add_inverted=add_inverted,
                    attributes=False,
                )
        if isinstance(index, Sequence):
            if attributes:
                return adapt_edge_iterable(
                    (
                        (v1,) + e_to
                        for edges_to, v1 in zip(index, itertools.count())
                        for e_to in edges_to
                    ),
                    add_inverted=add_inverted,
                    attributes=True,
                )
            else:
                return adapt_edge_iterable(
                    (
                        (v1, e_to)
                        for edges_to, v1 in zip(index, itertools.count())
                        for e_to in edges_to
                    ),
                    add_inverted=add_inverted,
                    attributes=False,
                )
        raise ValueError("graph must be Mapping or Sequence")

    def get(vertex: Any, _: Any) -> Union[Any, Sequence[Any]]:
        try:
            return index[vertex]
        except KeyError:
            return ()
        except IndexError:
            return ()

    return get


@overload
def adapt_edge_iterable(
    edges: Iterable[AnyFullEdge[T_vertex, T_weight, T_labels]],
    *,
    add_inverted: bool,
    attributes: Literal[False],
) -> Callable[[T_vertex, Any], Iterable[T_vertex]]:
    ...


@overload
def adapt_edge_iterable(
    edges: Iterable[WeightedOrLabeledFullEdge[T_vertex, T_weight, T_labels]],
    *,
    add_inverted: bool,
    attributes: Literal[True],
) -> Callable[[T_vertex, Any], Iterable[OutEdge[T_vertex, T_weight, T_labels]]]:
    ...


def adapt_edge_iterable(
    edges: Iterable[Sequence], *, add_inverted: bool, attributes: bool
) -> Callable:
    """
    Read a graph from an Iterable of edges and provide a neighbor function
    (`NextVertices` or `NextEdges`) from that data. Typically only used for test
    purposes.

    :param edges: The edges of your graph, each as Sequence (start_vertex, end_vertex,
        optional_attributes...), where edge data can be either a weight, or labels,
        or both (see `WeightedOrLabeledFullEdge`) or none
        (see `AnyFullEdge`).

    :param attributes: If set to True, the resulting neighbor function will yield
        edges with edge attributes for a given vertex and
        can be used as `NextEdges` function, but the edges of the given graph
        need to have attributes (`WeightedOrLabeledFullEdge`, not
        `AnyFullEdge`).

        If set to False, it will yield neighbor vertices, and can be
        used as `NextVertices` function, and edge data that your edges
        may contain will be ignored.

    :param add_inverted: If False, the resulting neighbor function will only yield
        for *edges*. If True, the function will also provide the
        same edges inverted (start and end vertex exchanged).
        Note: when this option is used, a
        copy of your graph will be held by the returned neighbor function.

    :return: Neighbor function that can be used as parameter for one of the traversal
        algorithms. See `OutEdge <outgoing_edges>` for the case of attributes.
    """
    edge_dict: dict[Any, list[Any]] = collections.defaultdict(list)
    if add_inverted:
        if attributes:
            # Labeled edges are provided and all data should be used
            for from_vertex, to_vertex, *data in edges:
                edge_dict[from_vertex].append((to_vertex, *data))
                edge_dict[to_vertex].append((from_vertex, *data))
            return adapt_edge_index(edge_dict, add_inverted=False, attributes=True)
        else:
            # Only start and end vertices of the edge should be used. Optional
            # additional data is ignored.
            for from_vertex, to_vertex, *data in edges:
                edge_dict[from_vertex].append(to_vertex)
                edge_dict[to_vertex].append(from_vertex)
            return adapt_edge_index(edge_dict, add_inverted=False, attributes=False)
    else:
        if attributes:
            for edge in edges:
                from_vertex, rest = edge[0], edge[1:]
                edge_dict[from_vertex].append(rest)
            return adapt_edge_index(edge_dict, add_inverted=False, attributes=True)
        else:
            for edge in edges:
                from_vertex, to_vertex = edge[0], edge[1]
                edge_dict[from_vertex].append(to_vertex)
            return adapt_edge_index(edge_dict, add_inverted=False, attributes=False)



from collections.abc import Mapping, Iterator
from typing import Callable, Union, Iterable
import re
import textwrap

from nographs import T_vertex, T_weight  # types
from nographs import traveling_salesman  # ._tsp
from nographs import GearDefault  # NOQA F401 (import needed by doc tests)  # ._gears

# noinspection PyProtectedMember
from nographs._extra_tsp import (  # NOQA F401 (import needed by doc tests)
    _traveling_salesman_int_vertices,
    GettableProto,
)


class GettableProtoTest:
    """
    >>> GettableProto.__getitem__(None, None)
    Traceback (most recent call last):
    NotImplementedError
    """


class TspTestsGeneral:
    """
    Unit tests for TSP

    >>> def test(edges, longest=False, vertices=None):
    ...    if vertices is None:
    ...        vertices=range(len(edges))
    ...    l, p = traveling_salesman(vertices, edges, longest)
    ...    print(l, list(p))

    -- TSP problem is not well-defined --

    Graph has no vertices and no edges
    >>> test(dict(), False)
    Traceback (most recent call last):
    RuntimeError: A TSP needs to have at least two vertices.

    Graph has one vertex and no edges
    >>> test({0: dict()}, False)
    Traceback (most recent call last):
    RuntimeError: A TSP needs to have at least two vertices.

    Graph has two vertices and one edge, but the relevant vertices are less than two
    >>> test({0: {1:1}, 1: {0:1}}, False, ())
    Traceback (most recent call last):
    RuntimeError: A TSP needs to have at least two vertices.
    >>> test({0: {1:1}, 1: {0:1}}, False, (1,))
    Traceback (most recent call last):
    RuntimeError: A TSP needs to have at least two vertices.

    -- TSP problem has no solution --

    A vertex has no outgoing edges
    >>> test({0: {1:1}}, False, (0, 1))
    Traceback (most recent call last):
    KeyError: 'No solution for the TSP exists: Some vertices have no outgoing edges.'

    A vertex has no incoming edges
    >>> test({0: {2:1}, 1: {2:1}, 2: {1:1}})
    Traceback (most recent call last):
    KeyError: 'No solution for the TSP exists: Some vertices have no incoming edges.'

    One of the vertices has only a self loop - but this will be ignored
    >>> test({0: {1:1}, 1: {0:1}, 2: {2:1},}, False)
    Traceback (most recent call last):
    KeyError: 'No solution for the TSP exists: Some vertices have no outgoing edges.'

    Graph has two vertices and one edge, but the relevant vertices are others
    >>> test({0: {1:1}, 1: {0:1}}, False, (0, 3))
    Traceback (most recent call last):
    KeyError: 'No solution for the TSP exists: Some vertices have no outgoing edges.'

    Graph has two isolated loops -> Unsolvable
    >>> test({0: {1:1}, 1: {0:1}, 2: {3:1}, 3: {2:1}}, False)
    Traceback (most recent call last):
    KeyError: 'No solution for the TSP exists.'

    Graph has two vertices and two suitable edges -> return result
    >>> test({0: {1:1}, 1: {0:1}}, False)
    2 [0, 1, 0]
    >>> test({0: {1:1}, 1: {0:1}}, True)
    2 [0, 1, 0]

    Three vertices, all connected, one edge direction shorter -> take this edge
    >>> test({0: {1:0, 2:1}, 1: {0:1, 2:1}, 2: {0:1, 1:1}}, False)
    2 [0, 1, 2, 0]
    >>> test({0: {1:1, 2:0}, 1: {0:1, 2:1}, 2: {0:1, 1:1}}, False)
    2 [0, 2, 1, 0]

    Three vertices, all connected, one edge direction longer, longest -> take this edge
    >>> test({0: {1:5, 2:1}, 1: {0:1, 2:1}, 2: {0:1, 1:1}}, True)
    7 [0, 1, 2, 0]
    >>> test({0: {1:1, 2:5}, 1: {0:1, 2:1}, 2: {0:1, 1:1}}, True)
    7 [0, 2, 1, 0]

    Four vertices, all connected, two consecutive edges shorter, shortest -> take this
    >>> test({0: {1:1, 2:0, 3:1}, 1: {0:1, 2:1, 3:1},
    ...       2: {0:1, 1:0, 3:1}, 3: {0:1, 1:1, 2:1}}, False)
    2 [0, 2, 1, 3, 0]

    Four vertices, all connected, two consecutive edges longer, longest -> take this
    >>> test({0: {1:1, 2:5, 3:1}, 1: {0:1, 2:1, 3:1},
    ...       2: {0:1, 1:5, 3:1}, 3: {0:1, 1:1, 2:1}}, True)
    12 [0, 2, 1, 3, 0]

    Four vertices, all connected, two relevant vertices -> solve this
    >>> test({0: {1:1, 2:1, 3:1}, 1: {0:1, 2:1, 3:1},
    ...       2: {0:1, 1:1, 3:1}, 3: {0:1, 1:1, 2:1}}, False, (1, 2))
    2 [1, 2, 1]
    """


class TspTestsOnlyForIntVertices:
    """
    >>> g = ((1, 1), (1,))
    >>> l, p = _traveling_salesman_int_vertices(g, GearDefault(), False)
    Traceback (most recent call last):
    RuntimeError: For vertex 1, 1 edges instead of 2 edges are given.
    """


# Source:
# TSPLIB – Sammlung von Benchmark-Instanzen des TSP und verschiedener Varianten
# http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/

ATSP_br17 = """
NAME:  br17
TYPE: ATSP
COMMENT: 17 city problem (Repetto)
DIMENSION:  17
EDGE_WEIGHT_TYPE: EXPLICIT
EDGE_WEIGHT_FORMAT: FULL_MATRIX
EDGE_WEIGHT_SECTION
9999    3    5   48   48    8    8    5    5    3    3    0    3    5    8    8    5
3 9999    3   48   48    8    8    5    5    0    0    3    0    3    8    8    5
5    3 9999   72   72   48   48   24   24    3    3    5    3    0   48   48   24
48   48   74 9999    0    6    6   12   12   48   48   48   48   74    6    6   12
48   48   74    0 9999    6    6   12   12   48   48   48   48   74    6    6   12
8    8   50    6    6 9999    0    8    8    8    8    8    8   50    0    0    8
8    8   50    6    6    0 9999    8    8    8    8    8    8   50    0    0    8
5    5   26   12   12    8    8 9999    0    5    5    5    5   26    8    8    0
5    5   26   12   12    8    8    0 9999    5    5    5    5   26    8    8    0
3    0    3   48   48    8    8    5    5 9999    0    3    0    3    8    8    5
3    0    3   48   48    8    8    5    5    0 9999    3    0    3    8    8    5
0    3    5   48   48    8    8    5    5    3    3 9999    3    5    8    8    5
3    0    3   48   48    8    8    5    5    0    0    3 9999    3    8    8    5
5    3    0   72   72   48   48   24   24    3    3    5    3 9999   48   48   24
8    8   50    6    6    0    0    8    8    8    8    8    8   50 9999    0    8
8    8   50    6    6    0    0    8    8    8    8    8    8   50    0 9999    8
5    5   26   12   12    8    8    0    0    5    5    5    5   26    8    8 9999
EOF
"""

TSP_gr_21 = """
NAME: gr21
TYPE: TSP
COMMENT: 21-city problem (Groetschel)
DIMENSION: 21
EDGE_WEIGHT_TYPE: EXPLICIT
EDGE_WEIGHT_FORMAT: LOWER_DIAG_ROW
EDGE_WEIGHT_SECTION
  0   510     0   635   355     0    91   415   605     0
385   585   390   350     0   155   475   495   120   240
  0   110   480   570    78   320    96     0   130   500
540    97   285    36    29     0   490   605   295   460
120   350   425   390     0   370   320   700   280   590
365   350   370   625     0   155   380   640    63   430
200   160   175   535   240     0    68   440   575    27
320    91    48    67   430   300    90     0   610   360
705   520   835   605   590   610   865   250   480   545
  0   655   235   585   555   750   615   625   645   775
285   515   585   190     0   480    81   435   380   575
440   455   465   600   245   345   415   295   170     0
265   480   420   235   125   125   200   165   230   475
310   205   715   650   475     0   255   440   755   235
650   370   320   350   680   150   175   265   400   435
385   485     0   450   270   625   345   660   430   420
440   690    77   310   380   180   215   190   545   225
  0   170   445   750   160   495   265   220   240   600
235   125   170   485   525   405   375    87   315     0
240   290   590   140   480   255   205   220   515   150
100   170   390   425   255   395   205   220   155     0
380   140   495   280   480   340   350   370   505   185
240   310   345   280   105   380   280   165   305   150
  0
EOF
"""


def read_tsp_problem(s: str) -> tuple[dict[str, str], dict[int, dict[int, int]]]:
    # generator that yields s line by line, each line without leading and trailing
    # spaces, first and last line are left out
    lines = (line.strip() for line in s[1:-1].split("\n"))

    # read header of TSP, extract size
    header_info = dict()
    while (line := next(lines)) != "EDGE_WEIGHT_SECTION":
        key, value = re.split(r": *", line)
        header_info[key] = value
    size = int(header_info["DIMENSION"])
    tsp_type = header_info["TYPE"]
    edge_weight_format = header_info["EDGE_WEIGHT_FORMAT"]

    numbers: Iterator[str]
    numbers = (word for line in lines for word in re.split(r"\W+", line))
    if edge_weight_format == "FULL_MATRIX":
        graph = {
            row: {column: int(next(numbers)) for column in range(size)}
            for row in range(size)
        }
    elif tsp_type == "TSP" and edge_weight_format == "LOWER_DIAG_ROW":
        numbers_tuple = tuple(numbers)
        # edges in forward direction
        numbers_iterator = iter(numbers_tuple)
        graph = {
            row: {column: int(next(numbers_iterator)) for column in range(row + 1)}
            for row in range(size)
        }
        # edges in backward direction
        numbers_iterator = iter(numbers_tuple)
        for row in range(size):
            for column in range(row + 1):
                graph[column][row] = int(next(numbers_iterator))
        numbers = numbers_iterator
    else:
        raise RuntimeError()

    if (line := next(numbers).strip()) != "EOF":
        raise RuntimeError(
            "TSP not correctly read: EOF not found where expected, instead found: "
            + line
        )

    return header_info, graph


def solve(
    test_name: str,
    function: Callable[[], Union[tuple[int, Iterable], tuple[float, Iterable]]],
    correct_length: T_weight,
    graph: Mapping[T_vertex, Mapping[T_vertex, T_weight]],
    time_stats: bool = False,
    mem_stats: bool = False,
    print_only_on_error: bool = False,
) -> bool:
    if not print_only_on_error and len(test_name) > 0:
        print()
        print("Running test:", test_name)

    if mem_stats and time_stats:
        correct = solve(
            "",
            function,
            correct_length,
            graph,
            time_stats=True,
            print_only_on_error=print_only_on_error,
        )
        correct = correct and solve(
            "",
            function,
            correct_length,
            graph,
            mem_stats=True,
            print_only_on_error=print_only_on_error,
        )
        return correct

    if mem_stats:
        import tracemalloc

        tracemalloc.start()
        tracemalloc.reset_peak()  # reset peak of previous execution parts

    if time_stats:
        import gc
        import timeit

        gc.collect()
        start = timeit.default_timer()

    try:
        length, path_iter = function()
    except OverflowError as e:
        if print_only_on_error:
            print("Running test:", test_name)
        if time_stats:
            print("Time: None. Test aborted. Exception caught:")
        if mem_stats:
            print("Memory peak: None. Test aborted. Exception caught:")
        print("\n  ".join(textwrap.wrap(str(e), 64)))
        return True

    if time_stats:
        stop = timeit.default_timer()
        print("Time: ", stop - start)

    if mem_stats:
        mem_size, mem_peak = tracemalloc.get_traced_memory()
        tracemalloc.stop()
        print(f"Memory peak: {int(mem_peak):,}")

    path = tuple(path_iter)
    path_is_correct = path[0] == path[-1]
    zero = correct_length - correct_length
    path_length = zero
    for v, w in zip(path, path[1:]):
        if w not in graph[v].keys():
            path_is_correct = False
            path_length = zero
            break
        path_length += graph[v][w]

    if length != correct_length or length != path_length or not path_is_correct:
        if print_only_on_error:
            print("Running test:", test_name)
        print("Got vertex sequence", path)
        print("This is " + ("" if path_is_correct else "not ") + "a valid path.")
        print(
            f"Got nominal length {length}, real length is {path_length}. "
            + f"Expected path with length {correct_length}."
        )
        return False
    return True


if __name__ == "__main__":
    """Additional unit tests for TSP: One of the standard TSPLIB problems."""
    for tsp, shortest, longest in (
        # Excluded the first tsp problem here. It is not necessary for QS, and
        # only used for the separate performance analysis.
        # (TSP_gr_21, 2707, 10680),
        (ATSP_br17, 39, 445),
    ):
        header_info, graph = read_tsp_problem(tsp)
        tsp_name = header_info["NAME"]
        tsp_type = header_info["TYPE"]
        tsp_is_symmetric = tsp_type != "ATSP"
        tsp_text = "{} ({}sym.), ".format(tsp_name, "" if tsp_is_symmetric else "a")

        solve(
            tsp_text + "traveling_salesman, shortest",
            lambda: traveling_salesman(range(len(graph)), graph),
            shortest,
            graph,
            print_only_on_error=True,
        )

        solve(
            tsp_text + "traveling_salesman, longest",
            lambda: traveling_salesman(range(len(graph)), graph, find_longest=True),
            longest,
            graph,
            print_only_on_error=True,
        )



from collections.abc import Iterator, Iterable, Sequence
from typing import Any, Union, Optional, Protocol, TypeVar
from abc import abstractmethod
import functools
import operator

from ._types import (
    T_vertex,
    T_weight,
    vertex_as_id,
)
from ._gears import (
    Gear,
    GearDefault,
)
from ._bidir_search import (
    BSearchShortestPathFlex,
)


# -- Protocols --
T_key_contra = TypeVar("T_key_contra", contravariant=True)
T_value_co = TypeVar("T_value_co", covariant=True)


class GettableProto(Protocol[T_key_contra, T_value_co]):
    """
    Protocol for a subscriptable readable container.

    Examples:

    - A dict[str, float] is a GettableProto[str, float]
    - A tuple[str, ...] is a GettableProto[int, str]
    """

    @abstractmethod
    def __getitem__(self, item: T_key_contra) -> Optional[T_value_co]:
        """
        Get the value that is stored for key *item*.
        If the Gettable does not contain *item*, a KeyError or an IndexError might
        be raised, or None is returned.
        """
        raise NotImplementedError


# -- Support functions ---
# For dealing with negative weights and/or search for longest travel.
# Exported for use in tests.


def weight_changes_for_negative_weights(
    weights: Iterator[T_weight], zero: T_weight
) -> tuple[bool, T_weight, bool]:
    """
    Return the necessary weight_factor and weight_offset for solving a TSP for
    shortest paths, if the weights are not guaranteed to be non-negative by using a
    TSP algorithm for only non-negative graph

    For avoiding negative weights, add a constant to the weights - we choose the
    positive value of the most negative weight. When using the results, revert that
    change accordingly. Note: For each edge of the resulting path, the constant goes
    into the resulting total distance exactly once.

    :weights: Iterates all non-None weights of the graph
    :one: Neutral element of the multiplication within T_weight
    """
    negate_weights = False
    weight_offset = zero - min(zero, min((weight for weight in weights), default=zero))
    return negate_weights, weight_offset, weight_offset != zero


def weight_changes_for_find_longest(
    weights: Iterator[T_weight], zero: T_weight
) -> tuple[bool, T_weight]:
    """
    Returns the necessary weight_factor and weight_offset for solving a TSP for
    longest path, if the weights are not guaranteed to be non-negative by using a TSP
    algorithm for only non-negative graph

    For finding the longest path, negate the weight. In order to avoid negative
    weights, we do the same correction as in weight_changes_for_negative_weights,
    but for the negated weight.

    :weights: Iterates all non-None weights of the graph
    :one: Neutral element of the multiplication within T_weight
    """
    negate_weights = True
    weight_offset = zero - min(
        zero, min((zero - weight for weight in weights), default=zero)
    )
    return negate_weights, weight_offset


def undo_weight_changes_in_travel_length(
    length: T_weight,
    no_of_edges: int,
    zero: T_weight,
    negate_weights: bool,
    weight_offset: T_weight,
) -> T_weight:
    # Undo the addition of weight_offset, that affects the path length
    # no_of_edges times.
    # (We have no multiplication in T_weight, so we repeat a subtraction. But since we
    # do this only once per vertex in the path, the additional runtime does not matter.)
    for i in range(no_of_edges):
        length -= weight_offset
    # Undo the negation of the edge weights. We can do this in one step for the whole
    # path length.
    return zero - length if negate_weights else length


# -------- TSP solution with integer-bases sets --------

IntVertex = int


def _traveling_salesman_int_vertices(
    graph: Sequence[Sequence[Optional[T_weight]]],
    gear: Gear[int, int, T_weight, Any],
    find_longest: bool = False,
) -> tuple[T_weight, Iterator[IntVertex]]:
    """
    Like traveling_salesman_flex, but the implementation is optimized for the
    case of integer vertices, and a graph needs to be a sequence of sequences of
    weights that forms a quadratic matrix:

    - Nested mappings are not allowed.
    - The matrix needs to be quadratic (vertices cannot be "left out", neither
      as source vertex of edges nor as target vertex of edges).
    - It is not possible to define a subset of vertices that the TSP tour need
      to use - automatically, all vertices will be visited.
    - If there is no edge between two vertices v and w, graph[v][w] needs to return
      None.

    :param graph: For each pair of vertices *v* and *w*, *graph* [*v*][*w*] is the
        weight of the edge from *v* to *w*.
    :param gear: The `gear` to be used internally (where vertices and vertex ids are
        represented by integers).
    :param find_longest: True means, find path with longest instead of shortest
        total distance.
    :return: Length of the found travel, and an iterator for the path of vertices
    :rtype: tuple[T_weight, Iterator[IntVertex]]
    """
    # Above, the rtype is documented manually because Sphinx autodocs with
    # option autodoc_typehints = 'description' cannot correctly
    # evaluate the type parameters of tuple (whilst typing.Tuple works).

    zero = gear.zero()

    # Solve TSPs with negative weights or with search for longest travels by using
    # TSP solution for positive weights, where the weights are manipulated in a
    # specific way, and the found travel length is corrected for that afterwards
    iter_weights = (weight for edges in graph for weight in edges if weight is not None)
    if find_longest:
        negate_weights, weight_offset = weight_changes_for_find_longest(
            iter_weights, zero
        )
        weights_changed = True
    else:
        (
            negate_weights,
            weight_offset,
            weights_changed,
        ) = weight_changes_for_negative_weights(iter_weights, zero)

    """ Solve TSP for non-negative weights
    See example in the tutorial that explains the concept of the algorithms.

    Implementation notes for the following optimized implementation:
    For memory and runtime efficiency, a search graph vertex is represented just
    by an integer. This is done as follows:
    - Sets of base graph vertices, e.g. the set of remaining vertices, are represented
      by the bits of an integer
    - The search graph vertex (a tuple) is encoded as:
      vertex_id + set_of_remaining_vertices * no_of_vertices
    """

    # --- Compute data about given graph and check if graph is well-formed ---
    # Size
    no_of_vertices = len(graph)
    if no_of_vertices < 2:
        # A TSP solution for a graph with just one vertex could be defined as being
        # a self-loop (is present in the graph), because "TSP-solutions for n-vertex
        # graphs have n edges", or as being an empty path (because "the salesman
        # does not need to travel). We do not decide here and exclude this case.
        # And it is also unclear, what a TSP solution in an empty graph should be
        # (also the empty path?).
        raise RuntimeError("A TSP needs to have at least two vertices.")

    # Is matrix quadratic?
    for v, edges in enumerate(graph):
        if len(edges) != no_of_vertices:
            raise RuntimeError(
                f"For vertex {v}, {len(edges)} edges instead of "
                + f"{no_of_vertices} edges are given."
            )

    # Pack transformed edge weights in nested tuples
    graph_forwards = tuple(
        tuple(
            None
            if (weight := graph[v][w]) is None
            else (zero - weight if negate_weights else weight) + weight_offset
            for w in range(no_of_vertices)
        )
        for v in range(no_of_vertices)
    )

    # The same, but for for incoming (!) edges of some vertex
    graph_backwards = tuple(
        tuple(
            None
            if (weight := graph[w][v]) is None
            else (zero - weight if negate_weights else weight) + weight_offset
            for w in range(no_of_vertices)
        )
        for v in range(no_of_vertices)
    )

    # arbitrary vertex to start (and end) paths with:
    start_vertex = 0

    # --- Define start and goal state of search graph ---

    # Vertices to visit when at start: all vertices, without start vertex
    # (vertex set containing only start vertex, as bitmask)
    start_vertices = 1 << start_vertex
    all_vertices = (1 << no_of_vertices) - 1  # set (bit mask) of all vertices
    vertices_to_visit = all_vertices - start_vertices

    # Start and goal states
    # Each as tuple (Current vertex, vertices still to visit afterwards)
    # Each tuple is encoded as a single integer as follows.
    start_state = start_vertex + vertices_to_visit * no_of_vertices
    goal_state = start_vertex + 0 * no_of_vertices

    # --- Define edges of search graph, by functions (both directions) ---

    # Precompiled adjacency matrix: For each vertex, the set of adjacent vertices (as
    # bit mask) (Typically, the adjacency matrix of a TSP is a full matrix. But in
    # order to get a better performance for sparse adjacency, the adjacency matrix is
    # precompiled from the graph, to avoid unnecessary failing dict lookups.)
    adjacent_vertices_forwards = tuple(
        functools.reduce(
            operator.or_,
            (
                1 << to_vertex
                if from_vertex != to_vertex and edges_from_vertex[to_vertex] is not None
                else 0
                for to_vertex in range(no_of_vertices)
            ),
            0,
        )
        for from_vertex, edges_from_vertex in enumerate(graph_forwards)
    )
    if not all(adjacent_vertices_forwards):
        raise KeyError(
            "No solution for the TSP exists: " + "Some vertices have no outgoing edges."
        )
    adjacent_vertices_backwards = tuple(
        functools.reduce(
            operator.or_,
            (
                1 << to_vertex
                if from_vertex != to_vertex and edges_from_vertex[to_vertex] is not None
                else 0
                for to_vertex in range(no_of_vertices)
            ),
            0,
        )
        for from_vertex, edges_from_vertex in enumerate(graph_backwards)
    )
    if not all(adjacent_vertices_backwards):
        raise KeyError(
            "No solution for the TSP exists: " + "Some vertices have no incoming edges."
        )

    def next_edges_forwards(state: int, _: Any) -> Iterable[tuple[int, T_weight]]:
        # 1. Interpret search graph vertex in terms of original graph
        to_visit_forwards, vertex = divmod(state, no_of_vertices)

        # 2. Yield edges in the search graph based on all existing edges in the base
        # graph from the current vertex to one of the vertices not visited in the
        # current travel so far. When all vertices are visited in the travel,
        # go to start vertex.
        if to_visit_forwards == 0:
            # At the last step forwards, the start vertex can be visited
            to_visit_forwards = start_vertices
        to_visit_next = to_visit_forwards & adjacent_vertices_forwards[vertex]
        edges_from_here = graph_forwards[vertex]
        while to_visit_next:
            to_vertex = to_visit_next.bit_length() - 1  # highest vertex in set
            to_vertex_set = 1 << to_vertex  # vertex (bit)set with just that vertex
            # weight of the edge (from_vertex, to_vertex):
            # (cannot be None, since we only iterate the non-None weights)
            weight: T_weight = edges_from_here[to_vertex]  # type: ignore[assignment]
            vertices_remaining_to_visit = to_visit_forwards - to_vertex_set
            yield (to_vertex + vertices_remaining_to_visit * no_of_vertices, weight)
            to_visit_next -= to_vertex_set  # to_vertex has been done

    def next_edges_backwards(state: int, _: Any) -> Iterable[tuple[int, T_weight]]:
        # 1. Interpret search graph vertex in terms of original graph
        to_visit_forwards, vertex = divmod(state, no_of_vertices)

        # 2. Yield edges in the search graph based on all existing edges in the base
        # graph to the current vertex from one of the vertices not visited in the
        # current travel so far. When all vertices are visited in the travel,
        # go to start vertex.
        to_visit_next = all_vertices - to_visit_forwards
        if to_visit_next != start_vertices:
            # Only at last step backwards, the start vertex can be visited
            to_visit_next -= start_vertices
        to_visit_next &= adjacent_vertices_backwards[vertex]
        edges_to_here = graph_backwards[vertex]
        vertices_remaining_to_visit = (
            to_visit_forwards
            if vertex == start_vertex
            else to_visit_forwards + (1 << vertex)
        )
        while to_visit_next:
            from_vertex = to_visit_next.bit_length() - 1  # highest vertex in set
            from_vertex_set = 1 << from_vertex  # vertex (bit)set with just that vertex
            # weight of the edge (from_vertex, to_vertex):
            # (cannot be None, since we only iterate the non-None weights)
            weight: T_weight = edges_to_here[from_vertex]  # type: ignore[assignment]
            yield (from_vertex + vertices_remaining_to_visit * no_of_vertices, weight)
            to_visit_next -= from_vertex_set  # from_vertex has been done

    s = BSearchShortestPathFlex[int, int, T_weight, Any](
        vertex_as_id, gear, (next_edges_forwards, next_edges_backwards)
    )

    try:
        length, state_path = s.start_from((start_state, goal_state), build_path=True)
    except KeyError:
        raise KeyError("No solution for the TSP exists.")

    # for the found path, return weight sum and vertices
    # (remove sets of remaining vertices)
    vertex_path = (v1 % no_of_vertices for v1 in state_path.iter_vertices_from_start())
    if weights_changed:
        # correct the found travel length
        length = undo_weight_changes_in_travel_length(
            length, no_of_vertices, zero, negate_weights, weight_offset
        )
    return length, vertex_path


def traveling_salesman_flex(
    vertices: Iterable[T_vertex],
    weights: GettableProto[T_vertex, GettableProto[T_vertex, T_weight]],
    gear: Gear[int, int, T_weight, Any],
    find_longest: bool = False,
) -> tuple[T_weight, Iterator[T_vertex]]:
    """
    Solve the traveling salesmen problem in directed graphs, in a slightly
    more general and flexible form than that of typical implementations. Performance
    has been a design goal.

    Find the path with the shortest, resp. the longest,
    length (sum of edge weights) through the relevant vertices (at least two), where
    each of these vertices is visited exactly once, in the form of a loop.

    If no solution can be found, raise a KeyError. If the problem is not well-defined,
    raise a RuntimeError.

    Flexibility:

    - Vertices can be any object except for None.

    - Edge weights can be positive, negative or zero.

    - The graph does not need to be symmetric (symmetric means: for each edge
      (i, j, weight), an edge (j, i, weight) exists).

    - The graph does not need to be complete (complete means: from each vertex there
      is an edge to all other vertices) and the algorithms profits
      if the graph is not complete.

    - Self-loops (i.e., an edge from a vertex to itself) are allowed (and will be
      ignored)

    - Optionally, a longest path is searched, instead of a shortest.

    - Edges from and to irrelevant vertices are allowed (and will be ignored)

    The function is an implementation of the algorithm shown in the
    `tutorial <reduction_of_other_problems>`,
    but it is based on the bidirectional search `BSearchShortestPathFlex`
    instead of the unidirectional traversal `TraversalShortestPaths`. And internally,
    it works with bit arrays instead of vertex sets.

    Recipes for solving other, but similar problems:

    - Your graph has multi-edges (i.e., from two vertices v and w, there are
      several edges from v to w): Give only the edge with the lowest edge
      weight between such vertices v and w.

    - Your graph is undirected: Give each edge in both directions.

    - Vertices can be visited multiple times: Build the distance graph
      and give it as input instead of your original graph.
      The distance graph can be computed by applying the traversal strategy
      `TraversalShortestPaths` for each of the vertices. Then, each tuple
      (start_vertex, reached_vertex, distance) is an edge of the distance
      graph. Save these edges, and the respective paths. When the TSP
      is solved on the distance graph, replace each edge in the result
      by the respective path of the original graph.

    :param vertices: The computed circle need to go through each of them exactly
        once. See `T_vertex`.
    :param weights: For each pair of vertices *v* and *w*, *weights* [*v*][*w*] is the
        weight of the edge from *v* to *w*, if the edge exists. Otherwise, a
        KeyError or an IndexError is raised, or None is returned. This allows
        for providing weights in nested tuples, lists or Mappings.
    :param gear: The `Gear` to be used internally (where vertices and vertex ids are
        represented by integers).
    :param find_longest: True means, find path with the longest instead of the
        shortest total distance
    :return: Length of the found travel, and an iterator for the path of vertices.
    :rtype: tuple[T_weight, Iterator[T_vertex]]
    """
    # See above why the rtype is documented manually

    vertex_tuple = tuple(vertices)
    no_of_vertices = len(vertex_tuple)

    # Generate quadratic matrix of weights
    int_graph = list[list[Optional[T_weight]]](
        [None] * no_of_vertices for v in range(no_of_vertices)
    )
    edge_to_vertex = [False] * no_of_vertices
    for v_int, v in enumerate(vertex_tuple):
        try:
            edges_from_v = weights[v]
        except (IndexError, KeyError):
            continue
        if edges_from_v is not None:
            for w_int, w in enumerate(vertex_tuple):
                try:
                    weight = edges_from_v[w]
                except (IndexError, KeyError):
                    continue
                if weight is not None:
                    int_graph[v_int][w_int] = weight
                    edge_to_vertex[w_int] = True

    # Solve TSP by using TSP solution for graphs with integer vertices
    length, path_raw = _traveling_salesman_int_vertices(int_graph, gear, find_longest)

    # Undo replacement of vertices by integers in the found path
    path = (vertex_tuple[v] for v in path_raw)

    return length, path


def traveling_salesman(
    vertices: Iterable[T_vertex],
    weights: GettableProto[T_vertex, GettableProto[T_vertex, T_weight]],
    find_longest: bool = False,
) -> tuple[Union[T_weight, float], Iterator[T_vertex]]:
    """
    Like `traveling_salesman_flex`, see there,
    but the gear is `GearDefault`, see also there.

    Note: As usual with GearDefault, the type for weight results (here: the path
    length) is extended by float. In case of `traveling_salesman`, a float will
    never be returned, if your graph does not contain float edge weights and the
    arithmetic operations on edge weights that are used (see `T_weight`) do not
    return floats. The reason is, that unsolvable TSP problems raise a KeyError,
    instead of returning a infinite path length.
    Use `traveling_salesman_flex` and choose another `Gear` if you like to have
    a stricter type specification (see also tutorial section `typing`).

    :param vertices: The computed circle need to go through each of them exactly
        once. See `T_vertex`.
    :param weights: For each pair of vertices *v* and *w*, *weights* [*v*][*w*] is the
        weight of the edge from *v* to *w*, if the edge exists. Otherwise, a
        KeyError or an IndexError is raised, or None is returned. This allows
        for providing weights in nested tuples, lists or Mappings.
    :param find_longest: True means, find path with the longest instead of the
        shortest total distance
    :return: Length of the found travel, and an iterator for the path of vertices.
    :rtype: tuple[Union[T_weight, float], Iterator[T_vertex]]
    """
    # See above why the rtype is documented manually

    # The following is a Gear[int, int, Union[T_weight, float], Any]
    gear = GearDefault[int, int, T_weight, Any]()
    return traveling_salesman_flex(vertices, weights, gear, find_longest=find_longest)



import nographs as nog  # noqa: F401 (used in doctests, undetected by flake 8)


class EdgeGadgets:
    """
    # ------------ adapt_edge_index -------------
    >>> # Weighted directed edges given as mapping:
    >>> edges = {
    ...    'A': (('B', 1), ('C', 2)),
    ...    'B': (('C', 5), ('D', 5)),
    ...    'C': (('D', 2), ('E', 1), ('F', 5)),
    ...    'D': (('F', 2), ('G', 5)),
    ...    'E': (('F', 5), ('B', 1)),
    ...    'F': (('G', 2),)
    ... }
    >>> import nographs as nog
    >>> next_vertices = nog.adapt_edge_index(edges,add_inverted=False,attributes=True)
    >>> traversal = nog.TraversalShortestPaths(next_vertices)
    >>> vertex = traversal.start_from("A", build_paths=True).go_to("G")
    >>> print("vertex:", vertex, "distance:", traversal.distance,
    ...       "path edge count:", traversal.depth, "path:", traversal.paths[vertex])
    vertex: G distance: 8 path edge count: 4 path: ('A', 'C', 'D', 'F', 'G')
    >>>
    >>> # The generated next_vertices function returns the empty list if no edges where
    >>> # given for it in the dict.
    >>> next_vertices("G", None)
    ()

    >>> # Same edges, given in a sequence, node id is integer index of the sequence
    >>> edges = (
    ...     ((1, 1), (2, 2)),
    ...     ((2, 5), (3, 5)),
    ...     ((3, 2), (4, 1), (5, 5)),
    ...     ((5, 2), (6, 5)),
    ...     ((5, 5), (1, 1)),
    ...     ((6, 2),)
    ... )
    >>> next_vertices = nog.adapt_edge_index(edges,add_inverted=False,attributes=True)
    >>> traversal = nog.TraversalShortestPaths(next_vertices)
    >>> vertex = traversal.start_from(0, build_paths=True).go_to(6)
    >>> print("vertex:", vertex, "distance:", traversal.distance,
    ...       "path edge count:", traversal.depth, "path:", traversal.paths[vertex])
    vertex: 6 distance: 8 path edge count: 4 path: (0, 2, 3, 5, 6)
    >>>
    >>> # The generated next_vertices function returns the empty list if the index is
    >>> # out of range
    >>> next_vertices(7, None)
    ()

    >>> # Access testing data, where weighted undirected edges are given as dict
    >>> edges = {
    ...     'A': (('B', 1), ('C', 2)),
    ...     'B': (('E', 1), ('D', 4), ('F', 1)),
    ...     'C': (('D', 1),),
    ...     'D': (('E', 1),),
    ...     'E': (('F', 2),),
    ...     'F': (('G', 2),)
    ... }
    >>> next_vertices = nog.adapt_edge_index(edges,add_inverted=True,attributes=True)
    >>> print({v: next_vertices(v, None) for v in "ABCDEF"}
    ... )  # doctest: +NORMALIZE_WHITESPACE
    {'A': [('B', 1), ('C', 2)], 'B': [('A', 1), ('E', 1), ('D', 4), ('F', 1)],
     'C': [('A', 2), ('D', 1)], 'D': [('B', 4), ('C', 1), ('E', 1)],
     'E': [('B', 1), ('D', 1), ('F', 2)], 'F': [('B', 1), ('E', 2), ('G', 2)]}

    >>> traversal = nog.TraversalMinimumSpanningTree(next_vertices)
    >>> [traversal.edge for vertex in traversal.start_from("A")
    ... ]  # doctest: +NORMALIZE_WHITESPACE
    [('A', 'B', 1), ('B', 'E', 1), ('B', 'F', 1), ('E', 'D', 1), ('D', 'C', 1),
    ('F', 'G', 2)]
    >>> # Note, that the minimum spanning tree contains edge ('D', 'C') in the direction
    >>> # not directly listed in the input.

    >>> # Access same testing graph, but the weighted undirected edges are given as
    >>> # list, vertices are int
    >>> edges = [
    ...     ((1, 1), (2, 2)),
    ...     ((4, 1), (3, 4), (5, 1)),
    ...     ((3, 1),),
    ...     ((4, 1),),
    ...     ((5, 2),),
    ...     ((6, 2),)
    ... ]
    >>> next_vertices = nog.adapt_edge_index(edges,add_inverted=True,attributes=True)
    >>> traversal = nog.TraversalMinimumSpanningTree(next_vertices)
    >>> [traversal.edge for vertex in traversal.start_from(0)]
    [(0, 1, 1), (1, 4, 1), (1, 5, 1), (4, 3, 1), (3, 2, 1), (5, 6, 2)]
    >>> # Note, that the minimum spanning tree contains edge (3, 2) in the direction
    >>> # not directly listed in the input.
    >>> next_vertices = nog.adapt_edge_index(set(),add_inverted=True,attributes=False)
    Traceback (most recent call last):
    ValueError: graph must be Mapping or Sequence

    # ------------ adapt_edge_iterable -------------
    >>> # Access testing data, where weighted directed edges are given as flat unsorted
    >>> # edge list
    >>> edges = [
    ...     ('A', 'B', 1), ('A', 'C', 2),
    ...     ('B', 'C', 5), ('B', 'D', 5),
    ...     ('C', 'D', 2), ('C', 'E', 1), ('C', 'F', 5),
    ...     ('D', 'F', 2), ('D', 'G', 5),
    ...     ('E', 'F', 5), ('E', 'B', 1),
    ...     ('F', 'G', 2)
    ... ]
    >>> import nographs as nog
    >>> next_vertices = nog.adapt_edge_iterable(
    ...     edges, add_inverted=False, attributes=True)
    >>> traversal = nog.TraversalShortestPaths(next_vertices)
    >>> vertex = traversal.start_from("A", build_paths=True).go_to("G")
    >>> print("vertex:", vertex, "distance:", traversal.distance,
    ...       "path edge count:", traversal.depth, "path:", traversal.paths[vertex])
    vertex: G distance: 8 path edge count: 4 path: ('A', 'C', 'D', 'F', 'G')
    >>>
    >>> # Access testing data, where weighted undirected edges are given as flat edge
    >>> # list
    >>> edges = [
    ...     ('A', 'B', 1), ('A', 'C', 2),
    ...     ('B', 'E', 1), ('B', 'D', 4), ('B', 'F', 1),
    ...     ('C', 'D', 1),
    ...     ('D', 'E', 1),
    ...     ('E', 'F', 2),
    ...     ('F', 'G', 2)
    ... ]
    >>> next_vertices = nog.adapt_edge_iterable(edges,add_inverted=True,attributes=True)
    >>> traversal = nog.TraversalMinimumSpanningTree(next_vertices)
    >>> [traversal.edge for vertex in traversal.start_from("A")
    ... ]  # doctest: +NORMALIZE_WHITESPACE
    [('A', 'B', 1), ('B', 'E', 1), ('B', 'F', 1), ('E', 'D', 1), ('D', 'C', 1),
    ('F', 'G', 2)]
    >>> # Note, that the minimum spanning tree contains edge ('D', 'C') in the
    >>> # direction not directly listed in the input.

    """



import nographs as nog  # noqa: F401 (used only by doctests)


class PathHandling:
    # noinspection PyShadowingNames
    """Special cases not covered in regular use

    >>> # noinspection PyProtectedMember
    >>> _path = nog._path

    >>> # noinspection PyProtectedMember
    >>> _take_except_for_last = _path._take_except_for_last
    >>> for iter_content_len in range(4):
    ...     list(_take_except_for_last(iter(range(iter_content_len))))
    []
    []
    [0]
    [0, 1]
    """



import sys
from decimal import Decimal, getcontext
from collections.abc import Iterator
from typing import Any, TypeVar, Union, TYPE_CHECKING

if sys.version_info >= (3, 11):
    from typing import assert_type
import unittest

from mpmath import mp, mpf  # type: ignore

import nographs as nog

# noinspection PyProtectedMember
from nographs._extra_tsp import GettableProto


# --- Types ---

T_any_typical_weight_type = TypeVar("T_any_typical_weight_type", float, Decimal, mpf)


# --- Tests ---


class TestWithTypes(unittest.TestCase):
    def test_variable_edge_weights(self) -> None:
        def test_with_small_weights(
            zero: T_any_typical_weight_type,
            one_half: T_any_typical_weight_type,
            one: T_any_typical_weight_type,
        ) -> int:
            def next_edges(
                i: int, _: Any
            ) -> Iterator[tuple[int, T_any_typical_weight_type]]:
                yield i + 1, one_half**i

            traversal = nog.TraversalShortestPaths[int, T_any_typical_weight_type, Any](
                next_edges
            )

            goal_difference = one_half**64
            previous_distance = zero
            for vertex in traversal.start_from(1):
                distance = traversal.distance
                assert isinstance(distance, type(one))
                if one - distance == goal_difference:
                    return vertex
                if distance == previous_distance:
                    raise RuntimeError("Error: Distance stays constant")
                previous_distance = distance
            raise RuntimeError("Unexpected end of loop")  # pragma: no cover

        with self.assertRaises(RuntimeError) as cm:
            test_with_small_weights(float(0), float("0.5"), float(1))
        self.assertEqual(cm.exception.args, ("Error: Distance stays constant",))

        getcontext().prec = 75  # precision (number of places)
        self.assertEqual(
            test_with_small_weights(Decimal(0), Decimal("0.5"), Decimal(1)), 65
        )

        mp.prec = 64
        self.assertEqual(test_with_small_weights(mpf(0), mpf("0.5"), mpf(1)), 65)

    def test_traversal_typing_docs_example(self) -> None:
        def next_edges(i: int, _: Any) -> Iterator[tuple[int, int]]:
            j = (i + i // 6) % 6
            yield i + 1, j * 2 + 1
            if i % 2 == 0:
                yield i + 6, 7 - j
            elif i % 1200000 > 5:
                yield i - 6, 1

        traversal = nog.TraversalShortestPaths[int, int, Any](next_edges)
        v = traversal.start_from(0, build_paths=True).go_to(5)
        d = traversal.distance
        p = tuple(traversal.paths.iter_vertices_from_start(v))
        if sys.version_info >= (3, 11):
            if TYPE_CHECKING:
                # for Python>=3.11, check types
                assert_type(v, int)
                assert_type(d, Union[int, float])
                assert_type(p, tuple[int, ...])
        else:
            reveal_type = print  # will never be used; just for IDEs
            # for Python<3.11, reveal types
            if TYPE_CHECKING:
                reveal_type(v)  # reveals: int
                reveal_type(d)  # reveals: Union[int, float]
                reveal_type(p)  # reveals: tuple[int, ...]
        self.assertEqual(v, 5)
        self.assertEqual(d, 24)
        self.assertEqual(p, (0, 1, 2, 3, 4, 10, 16, 17, 11, 5))

    def test_tsp_typing_docs_example(self) -> None:
        """The MyPy run will detect if the typing goes wrong here.

        The NOQA tells flake8 that it is on purpose that g1 and g2 are
        assigned but never used.
        """
        g1: GettableProto[int, str] = dict[int, str]([(0, "a"), (1, "b")])  # NOQA F841
        g2: GettableProto[int, str] = ["a", "b"]  # NOQA F841
